/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry;


import net.rim.device.api.system.PersistentObject;
import net.rim.device.api.system.PersistentStore;

import com.funambol.util.Log;
import com.funambol.util.StaticDataHelper;

/**
 * Configuration class for configuration details
 */
public class Configuration
{
    // ------------------------------------------------------------ Constants
    public static final String VERSION_KEY              = "VERSION";
    public static final String LOG_LEVEL_KEY            = "LOG_LEVEL";
    public static final String SYNC_URL_KEY             = "SYNC_URL";
    public static final String USERNAME_KEY             = "USERNAME";
    public static final String PASSWORD_KEY             = "PASSWORD";
    public static final String SYNC_CONTACT_FLAG_KEY    = "SYNC_CONTACT";
    public static final String SYNC_CONTACT_URI_KEY     = "CONTACT_URI";
    public static final String SYNC_CAL_FLAG_KEY        = "SYNC_CAL";
    public static final String SYNC_CAL_URI_KEY         = "CAL_URI";
    public static final String SYNC_NOTE_FLAG_KEY       = "SYNC_NOTE";
    public static final String SYNC_NOTE_URI_KEY        = "NOTE_URI";
    public static final String SYNC_TASK_FLAG_KEY       = "SYNC_TASK";
    public static final String SYNC_TASK_URI_KEY        = "TASK_URI";
    public static final String POLL_PIM_FLAG_KEY        = "POLL_PIM";
    public static final String POLL_PIM_TIME_KEY        = "POLL_PIM_TIME";
    public static final String RANGE_PAST_KEY           = "RANGE_PAST";
    public static final String RANGE_FUTURE_KEY         = "RANGE_FUTURE";
    public static final String SYNC_TYPE_KEY			= "SYNC_TYPE";

    public static final String      VERSION                 = "20060721";
    public static String            MAXMSGSIZE              = "16000";
    public static String            PROTOCOL;
    public static String            RETRY;
    public static int               logLevel                = 0; // disabled on plugin, error on api;
    
    public static final int         CONF_INVALID            = -2;
    public static final int         CONF_NOTSET             = -1;
    public static final int         CONF_OK                 = 0;        
    public static final int         PARAM_NOTSET            = 1;

    public static String            syncUrl;
    public static String            calendarSourceUri;
    public static String            contactSourceUri;
    public static String            noteSourceUri;
    public static String            taskSourceUri;

    public static boolean           enablePollingPIM;
    public static int               pollingPIMInterval;
    
    public static int               rangePast;
    public static int               rangeFuture;

    public static String            username;
    public static String            password;
    
    public static int				syncType;
    
    public static boolean           calendarSyncAllowed = true;
    public static boolean           syncCalendar;
    public static boolean           contactSyncAllowed = true;
    public static boolean           syncContact;
    public static boolean           noteSyncAllowed = true;
    public static boolean           syncNote;
    public static boolean           taskSyncAllowed = true;
    public static boolean           syncTask;

    public static boolean           mailSyncAllowed = false;
    public static boolean           syncMail;
    
    private static void copyDefaults()
    {
        syncUrl = StaticDataHelper.getServerUriDefault();
        username = StaticDataHelper.getUserDefault();
        password = StaticDataHelper.getPasswordDefault();
        
        syncContact = StaticDataHelper.getSyncContactDefault();
        contactSourceUri = StaticDataHelper.getContactSourceUriDefault();
        
        syncCalendar = StaticDataHelper.getSyncCalendarDefault();
        calendarSourceUri = StaticDataHelper.getCalendarSourceUriDefault();
        
        syncTask = StaticDataHelper.getSyncTaskDefault();
        taskSourceUri = StaticDataHelper.getTaskSourceUriDefault();
        
        syncNote = StaticDataHelper.getSyncNoteDefault();
        noteSourceUri = StaticDataHelper.getNoteSourceUriDefault();
        
        enablePollingPIM = StaticDataHelper.getEnablePollingPIMDefault();
        pollingPIMInterval = StaticDataHelper.getPollingPIMIntervalDefault();
        
        rangePast = StaticDataHelper.getRangePastDefault();
        rangeFuture = StaticDataHelper.getRangeFutureDefault();

        syncType = StaticDataHelper.getSyncTypeDefault();
    }

    /**
     * This must be called once at the beginning of the application.
     **/
    public static void init(Class c) throws Exception
    {
        StaticDataHelper.init(c,StaticDataHelper.XML_LANGUAGE);
        copyDefaults();
    }
    // ------------------------------------------------------------ Public
    /**
     * Load the current config from the persistent store.
     * @return:
     * <li><b>CONF_OK</b>: if all the data were present in the store.
     * <li><b>CONF_NOTSET</b>: if the store is not present. Configuration
     * remains untouched.
     * <li><b>CONF_INVALID</b>: if the store does not contain valid data.
     * Configuration is reverted to default.
     * Note: if a parameter is not present in the store, the current value is
     * kept for it.
     */
    public static int load()
    {
        Log.info("Loading Configuration data from persistent storage");
        final PersistentObject pobj = PersistentStore.getPersistentObject(SyncClient.CONFIG_KEY);
        try
        {
            final ConfigData contents = (ConfigData) pobj.getContents();
            if (contents == null)
            {
                    Log.info("Configuration not set, using current values.");
                    return -1;
            }
            if (contents.get(Configuration.VERSION_KEY) == null)
            {
                    Log.error("Configuration invalid: reverting to default.");
                    PersistentStore.destroyPersistentObject(SyncClient.CONFIG_KEY);
                    copyDefaults();
                    return -2;
            }

            logLevel = contents.getValue(Configuration.LOG_LEVEL_KEY, logLevel);
            
            syncUrl = contents.getValue(Configuration.SYNC_URL_KEY, syncUrl);
            username = contents.getValue(Configuration.USERNAME_KEY, username);
            password = contents.getValue(Configuration.PASSWORD_KEY, password);
            
            syncContact = contents.getValue(Configuration.SYNC_CONTACT_FLAG_KEY, syncContact);
            contactSourceUri = contents.getValue(Configuration.SYNC_CONTACT_URI_KEY, contactSourceUri);
            
            syncCalendar = contents.getValue(Configuration.SYNC_CAL_FLAG_KEY, syncCalendar);
            calendarSourceUri = contents.getValue(Configuration.SYNC_CAL_URI_KEY, calendarSourceUri);
            
            syncTask = contents.getValue(Configuration.SYNC_TASK_FLAG_KEY, syncTask);
            taskSourceUri = contents.getValue(Configuration.SYNC_TASK_URI_KEY, taskSourceUri);
            
            syncNote = contents.getValue(Configuration.SYNC_NOTE_FLAG_KEY, syncNote);
            noteSourceUri = contents.getValue(Configuration.SYNC_NOTE_URI_KEY, noteSourceUri);
            
            enablePollingPIM = contents.getValue(Configuration.POLL_PIM_FLAG_KEY, enablePollingPIM);
            pollingPIMInterval = contents.getValue(Configuration.POLL_PIM_TIME_KEY, pollingPIMInterval);
            
            rangePast = contents.getValue(Configuration.RANGE_PAST_KEY, rangePast);
            rangeFuture = contents.getValue(Configuration.RANGE_FUTURE_KEY, rangeFuture);
           
            syncType = contents.getValue(Configuration.SYNC_TYPE_KEY, syncType);
            
            logLevel = contents.getValue(Configuration.LOG_LEVEL_KEY, logLevel);
            
            return CONF_OK;
        }
        catch (final ClassCastException cce)
        {
            Log.error("Configuration invalid: reverting to default.");
            PersistentStore.destroyPersistentObject(SyncClient.CONFIG_KEY);
            copyDefaults();
            return -2;
        }
    }

        
    /**
     * Save the current config to the persistent store.
     */
    public static int save(SyncClient app)
    {
        Log.info("Saving Configuration data to persistent storage");
        final PersistentObject pobj = PersistentStore.getPersistentObject(SyncClient.CONFIG_KEY);
        ConfigData contents = (ConfigData) pobj.getContents();
        if (contents == null)
        {
            contents = new ConfigData();
        }

        // This field can be used to check which version of the sw
        // saved the data. If not present, it is assumed it's an
        // old version and the saved config is discarded.
        contents.setValue(Configuration.VERSION_KEY, VERSION);
        contents.setValue(Configuration.LOG_LEVEL_KEY, logLevel);

        contents.setValue(Configuration.SYNC_URL_KEY, syncUrl);
        contents.setValue(Configuration.USERNAME_KEY, username);
        contents.setValue(Configuration.PASSWORD_KEY, password);
        
        contents.setValue(Configuration.SYNC_CONTACT_FLAG_KEY, syncContact);
        contents.setValue(Configuration.SYNC_CONTACT_URI_KEY, contactSourceUri);
        
        contents.setValue(Configuration.SYNC_CAL_FLAG_KEY, syncCalendar);
        contents.setValue(Configuration.SYNC_CAL_URI_KEY, calendarSourceUri);
        
        contents.setValue(Configuration.SYNC_TASK_FLAG_KEY, syncTask);
        contents.setValue(Configuration.SYNC_TASK_URI_KEY, taskSourceUri);
        
        contents.setValue(Configuration.SYNC_NOTE_FLAG_KEY, syncNote);
        contents.setValue(Configuration.SYNC_NOTE_URI_KEY, noteSourceUri);
        
        contents.setValue(Configuration.POLL_PIM_FLAG_KEY, enablePollingPIM);
        contents.setValue(Configuration.POLL_PIM_TIME_KEY, pollingPIMInterval);
        
        contents.setValue(Configuration.RANGE_PAST_KEY, rangePast);
        contents.setValue(Configuration.RANGE_FUTURE_KEY, rangeFuture);
        
        contents.setValue(Configuration.SYNC_TYPE_KEY, syncType);
        
        synchronized (pobj)
        {
            pobj.setContents(contents);
            pobj.commit();
            app.initAllSourceConfigs(false);
        }
        return CONF_OK;
    }

    
    public static int append(String key,byte[] value)
    {
        final PersistentObject pobj = PersistentStore.getPersistentObject(SyncClient.CONFIG_KEY);
        ConfigData contents = (ConfigData) pobj.getContents();
        if (contents == null)
        {
            contents = new ConfigData();
        }
        contents.setValue(key,value);
        synchronized (pobj)
        {
            pobj.setContents(contents);
            pobj.commit();
        }
        return CONF_OK;
    }
    

    public static byte[] getValue(String key,byte[] value)
    {
        final PersistentObject pobj = PersistentStore.getPersistentObject(SyncClient.CONFIG_KEY);
        try
        {
            final ConfigData contents = (ConfigData) pobj.getContents();
            if (contents == null)
            {
                return null;
            }
            return contents.getValue(key,value);
        }
        catch(final ClassCastException e)
        {}
        return null;
    }
}
