/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry;

import java.util.Hashtable;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;

import java.io.InputStream;
import javax.microedition.pim.PIM;
import javax.microedition.io.Connector;
import javax.microedition.io.HttpConnection;

import net.rim.blackberry.api.pdap.BlackBerryPIM;
import net.rim.blackberry.api.pdap.BlackBerryPIMList;
import net.rim.device.api.system.Application;
import net.rim.device.api.system.Bitmap;
import net.rim.device.api.system.RadioInfo;
import net.rim.device.api.ui.Font;
import net.rim.device.api.ui.FontFamily;
import net.rim.device.api.ui.Manager;
import net.rim.device.api.ui.UiApplication;
import net.rim.device.api.ui.component.Dialog;
import net.rim.device.api.ui.container.MainScreen;
import net.rim.device.api.system.DeviceInfo;
import net.rim.blackberry.api.browser.Browser;

import com.funambol.blackberry.helper.HelperFactory;
import com.funambol.blackberry.helper.PIMItemHelper;
import com.funambol.blackberry.source.BBPIMSyncSource;
import com.funambol.blackberry.source.Poller;
import com.funambol.blackberry.ui.*;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.spds.CompressedSyncException;
import com.funambol.syncml.spds.ConnectionDeniedException;
import com.funambol.syncml.spds.DeviceConfig;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncConfig;
import com.funambol.syncml.spds.SyncManager;
import com.funambol.syncml.spds.SyncException;
import com.funambol.util.BasicConnectionHandler;
import com.funambol.util.BlackberryHelper;
import com.funambol.util.Log;
import com.funambol.util.SimpleXMLParser;
import com.funambol.util.StaticDataHelper;


/**
 * Provides the user interface with the configuration settings, mail
 * synchronization capabilities (polling included), and contacts and calendar
 * events synchronization
 * 
 */
public class SyncClient extends UiApplication implements Runnable//,FolderListener,SendListener
{
    // To control custom features.  Disable for Funambol
    public static final boolean CUSTOMS_ENABLED = false;
    
    public static final int CUSTOMS_COMPANY_DIRECTORY   = 1;
    public static final int CUSTOMS_SYNC_RANGE          = 2; //Not used yet
    public static final int CUSTOMS_UPDATE              = 4;
    public static final int CUSTOMS_ROLLOVER            = 8;
    public static final int CUSTOMS_TITLE               = 16;

    public static final int CUSTOMS = 
        CUSTOMS_COMPANY_DIRECTORY +
        CUSTOMS_UPDATE +
        CUSTOMS_SYNC_RANGE + 
        CUSTOMS_ROLLOVER + 
        CUSTOMS_TITLE;

    public static final long            CONFIG_KEY                  = 0xbc565a75eda532eL;
    public static final long            CONTACT_STORAGE_KEY         = 0xbc565a75eda5330L;
    public static final long            CALENDAR_STORAGE_KEY        = 0xbc565a75eda5340L;
    public static final long            NOTE_STOREAGE_KEY           = 0xbc565a75eda5350L;
    public static final long            TASK_STORAGE_KEY            = 0xbc565a75eda5360L;
    public static final String          SYNC_CONTACT_CONFIG         = "SYNC_CONTACT_CONFIG";
    public static final String          SYNC_CAL_CONFIG             = "SYNC_CAL_CONFIG";
    public static final String          SYNC_NOTE_CONFIG            = "SYNC_NOTE_CONFIG";
    public static final String          SYNC_TASK_CONFIG            = "SYNC_TASK_CONFIG";
 
    //public static final String MAIL             = "mail";
    //public static final String BRIEFCASE        = "briefcase";

    public static final String CONTACT          = "contact";
    public static final String CONTACT_TYPE     = "text/x-s4j-sifc";
    public static final String CALENDAR         = "calendar";
    public static final String CALENDAR_TYPE    = "text/x-s4j-sife";
    public static final String NOTE             = "note";
    public static final String NOTE_TYPE        = "text/x-s4j-sifn";
    public static final String TASK             = "task";
    public static final String TASK_TYPE        = "text/x-s4j-sift";
    
    public static String                VER                         = ""; //Set on startup

    final private static String         ERROR_NOT_AUTHORIZED        = "401";
    final private static String         ERROR_NOT_FOUND             = "404";
    final private static String         ERROR_REFRESH_REQUIRED      = "508";
    final private static String         ERROR_SERVER                = "511";
    
    public static final String  MANUAL_SYNC = "manual";
    public static final String  SCHEDULED_SYNC = "scheduled";
    public static String syncType;
    private static int scheduledAttempt = 0;
    private static final int  NEW_POL_TIME = 5; //minutes
    private Poller pol;
    
    // Used to select which source to sync in synchronize()
    public static final int             SYNC_CONTACTS               = 1;
    public static final int             SYNC_CALENDAR               = 2;
    public static final int             SYNC_TASKS                  = 4;
    public static final int             SYNC_NOTES                  = 8;
    public static final int             SYNC_MAIL                   = 16;
    public static final int             SYNC_ALL                    = 255;
    public static final int             SYNC_SUPPORTED              = SYNC_CONTACTS + 
                                                                      SYNC_CALENDAR +
                                                                      SYNC_TASKS +
                                                                      SYNC_NOTES;
    
    public final static int             REFRESH_FROM_SERVER         = 1;
    public final static int             REFRESH_TO_SERVER           = 2;
    
    private DeviceConfig                devconf                     = null;
    private boolean                     doRefresh                   = false;                        
    private boolean                     doCancel                    = false;
    
    private int                         refreshDirection            = 0;
    
    private boolean                     updateFlag                  = false;
    private boolean                     updateRequired              = false;
    private String                      updateOTA                   = null;
    private String                      updateURL                   = null;
    
    private boolean                     inSynchronize               = false;
    
    private AppScreen                   appScreen                   = null;
    private Poller                      pollerPIM                   = null;
    private Thread                      thread                      = null;
    private Hashtable                   sources                     = null;
    private SourceConfig                contactSourceConfig;
    private SourceConfig                calendarSourceConfig;
    private SourceConfig                noteSourceConfig;
    private SourceConfig                taskSourceConfig;
    private int                         minCacheSize                = 10;
    private int                         maxCacheSize                = 100;

    private SyncConfig sc;
    
    private boolean                        displayConnectivityError    = false;
    
    private boolean                     compressionRetry            = false;
    
    private boolean enableSyncCalendar = true;
    
    static
    {
        StaticDataHelper.setClientVersion("Unknown");
    }
    
    /**
     * The default constructor. It opens internally the BlackBerry PIM lists
     * (contacts and calendar events), adds a listener to them and to the mail
     * folders, and pushes the home screen to allow user configuration
     */
    public SyncClient() throws Exception
    {
        // DEBUG only (needed to activate log during conf load
        Log.setLogLevel(Log.TRACE);
        /*
         * get configurations contents from the store. The class configuration
         * initializes also StaticDataHelper defaults.
         */
        Configuration.init(this.getClass());
        Configuration.load();
        sc = new SyncConfig();
        initAllSourceConfigs(true);
        
        Font font = Font.getDefault();
        try {
            FontFamily ff = null;
            ff = FontFamily.forName("Verdana");
            font = ff.getFont(StaticDataHelper.FONT_STYLE, StaticDataHelper.FONT_SIZE);
        } catch (ClassNotFoundException e) {}
        Font.setDefaultFont(font);
        
        // Set the log level
        Log.setLogLevel(Configuration.logLevel);

        // Create syncsource configs
        sources = new Hashtable();
        BlackBerryPIMList list;
        BBPIMSyncSource syncSource;

        StaticDataHelper.info("Creating Contact sync source");
        list = (BlackBerryPIMList) PIM.getInstance().openPIMList(PIM.CONTACT_LIST,PIM.READ_WRITE);
        syncSource = new BBPIMSyncSource(SyncClient.SYNC_CONTACTS,
                                contactSourceConfig
                                ,list
                                ,HelperFactory.getSIFContactHelper(Configuration.rangePast,Configuration.rangeFuture)
                                ,SyncClient.CONTACT_STORAGE_KEY
                                ,this
                                ,minCacheSize
                                ,maxCacheSize
                                );
        sources.put(new Integer(SyncClient.SYNC_CONTACTS), syncSource);

        StaticDataHelper.info("Creating Event sync source");
        list = (BlackBerryPIMList) PIM.getInstance().openPIMList(PIM.EVENT_LIST,PIM.READ_WRITE);
        syncSource = new BBPIMSyncSource(SyncClient.SYNC_CALENDAR,
                            calendarSourceConfig
                            ,list
                            ,HelperFactory.getSIFEventHelper(Configuration.rangePast,Configuration.rangeFuture)
                            ,SyncClient.CALENDAR_STORAGE_KEY
                            ,this
                            ,minCacheSize
                            ,maxCacheSize
                            );
        sources.put(new Integer(SyncClient.SYNC_CALENDAR), syncSource);

        StaticDataHelper.info("Creating Note sync source");
        list = (BlackBerryPIMList) PIM.getInstance().openPIMList(BlackBerryPIM.MEMO_LIST,PIM.READ_WRITE);
        syncSource = new BBPIMSyncSource(SyncClient.SYNC_NOTES,
                            noteSourceConfig
                            ,list
                            ,HelperFactory.getSIFNoteHelper(Configuration.rangePast,Configuration.rangeFuture)
                            ,SyncClient.NOTE_STOREAGE_KEY
                            ,this
                            ,minCacheSize
                            ,maxCacheSize
                            );
        sources.put(new Integer(SyncClient.SYNC_NOTES), syncSource);

        StaticDataHelper.info("Creating Task sync source");
        list = (BlackBerryPIMList) PIM.getInstance().openPIMList(PIM.TODO_LIST,PIM.READ_WRITE);
        PIMItemHelper helper = HelperFactory.getSIFTaskHelper(Configuration.rangePast,Configuration.rangeFuture);
        helper.setDeleteForUpdate(true);
        syncSource = new BBPIMSyncSource(SyncClient.SYNC_TASKS,
                            taskSourceConfig
                            ,list
                            ,helper
                            ,SyncClient.TASK_STORAGE_KEY
                            ,this
                            ,minCacheSize
                            ,maxCacheSize
                            );
        sources.put(new Integer(SyncClient.SYNC_TASKS), syncSource);
        helper = null;
        list = null;
        syncSource = null;

        // Turn on polling, if necessary
        if (Configuration.enablePollingPIM)
        {
            enablePollingPIM();
        }
        
        // Initalize the interfaces
        appScreen = new AppScreen(this);
        long lastSync = 
                Math.max(
                        Math.max(
                                calendarSourceConfig.getNextAnchor(),
                                contactSourceConfig.getNextAnchor()
                                ),
                        Math.max(
                                taskSourceConfig.getNextAnchor(),
                                noteSourceConfig.getNextAnchor()
                                )
                        );
        if (lastSync > 0)
        {
            appScreen.setLastSync(lastSync);
        }

        if (contactSourceConfig.getNextAnchor() == 0)
            appScreen.getSyncScreen().displayStatus(SYNC_CONTACTS, StaticDataHelper.getLanguage("sync_unsynchronized"));
        if (calendarSourceConfig.getNextAnchor() == 0)
            appScreen.getSyncScreen().displayStatus(SYNC_CALENDAR, StaticDataHelper.getLanguage("sync_unsynchronized"));
        if (taskSourceConfig.getNextAnchor() == 0)
            appScreen.getSyncScreen().displayStatus(SYNC_TASKS, StaticDataHelper.getLanguage("sync_unsynchronized"));
        if (noteSourceConfig.getNextAnchor() == 0)
            appScreen.getSyncScreen().displayStatus(SYNC_NOTES, StaticDataHelper.getLanguage("sync_unsynchronized"));
        
        
        /*
         * The UI engine must be accessed holding the event lock object shared
         * with the application's main thread, otherwise an
         * IllegalStateException is thrown (the event lock is held here where
         * the screen is pushed in the foreground)
         */
        synchronized(SyncClient.getEventLock())
        {
            pushScreen(appScreen);
        }
        
        if (SyncClient.checkCustom(SyncClient.CUSTOMS_UPDATE))
        {
            checkUpdate(true, false);
        }
    }


    /**
     * Must be called on initialization
     * 
     * @return  False on failure
     */
    public boolean loadVersion()
    {
        String version = BBVersion.VERSION;
        if (version.length() < 5)
        {
            return false;
        }
        SyncClient.VER = version;
        StaticDataHelper.setClientVersion(version);
        StaticDataHelper.info("Version: " + version);
        return true;
    }
    
    /**
     * Helper function to prompt the user for a yes/no answer
     * 
     * @param question
     * @param defaultyes
     * @return                          True on yes
     */
    public static boolean askYesNoQuestion(String question, boolean defaultyes)
    {
        synchronized (SyncClient.getEventLock())
        {
            return Dialog.YES == Dialog.ask(Dialog.D_YES_NO,question,defaultyes ? Dialog.YES : Dialog.NO);
        }
    }

    /**
     * Helper function to prompt to continue or cancel
     * 
     * @param message
     * @return  True for next
     */
    public static boolean promptNext(String message)
    {
        String [] buttons = new String [] {
                        "Next", "Cancel"
        };
        int [] values = new int [] {
                        0, 1
        };

        synchronized (SyncClient.getEventLock())
        {
                return 0 == Dialog.ask(message, buttons, values, 0);
        }
    }
    
    /**
     * Helper function to display a simple notice
     * 
     * @param message
     */
    public static void showMessage(String message)
    {
        final Dialog dialog = new Dialog(
                Dialog.D_OK,
                message,
                1,
                Bitmap.getPredefinedBitmap(Bitmap.EXCLAMATION),
                Manager.LEAVE_BLANK_SPACE);
        synchronized (SyncClient.getEventLock())
        {
            UiApplication.getUiApplication().pushScreen(dialog);
        }
    }

    /**
     * Push a screen to the forefront
     * 
     * @param scr       The screen to push
     */
    public void showScreen(MainScreen scr)
    {
        if (scr.isDisplayed())
        {
            hideScreen(scr);
        }
        synchronized (SyncClient.getEventLock())
        {
            UiApplication.getUiApplication().pushScreen(scr);
        }
    }

    /**
     * Pull a screen to the back
     * 
     * @param scr       The screen to hide
     */
    public void hideScreen(MainScreen scr)
    {
        if (!scr.isDisplayed())
        {
            return;
        }
        synchronized (SyncClient.getEventLock())
        {
            UiApplication.getUiApplication().popScreen(scr);
        }
    }

    /**
     * Initalize all syncsource configs.  
     * 
     * @param read      Pass true to read from the device persistent storage.  Forwarded to initSourceConfig
     */
    public void initAllSourceConfigs(boolean read)
    {
        long last = 0;
        long next = 0;
        boolean restore;
        
        
        restore = false;
        if (contactSourceConfig != null)
        {
            last = contactSourceConfig.getLastAnchor();
            next = contactSourceConfig.getNextAnchor();
            restore = true;
        }
        contactSourceConfig  = SyncClient.initSourceConfig(SyncClient.SYNC_CONTACT_CONFIG, CONTACT, CONTACT_TYPE, Configuration.contactSourceUri, read);
        if (restore)
        {
            contactSourceConfig.setLastAnchor(last);
            contactSourceConfig.setNextAnchor(next);
        }
        SyncClient.saveSourceConfig(SyncClient.SYNC_CONTACT_CONFIG, contactSourceConfig);
        
        
        restore = false;
        if (calendarSourceConfig != null)
        {
            last = calendarSourceConfig.getLastAnchor();
            next = calendarSourceConfig.getNextAnchor();
            restore = true;
        }
        calendarSourceConfig = SyncClient.initSourceConfig(SyncClient.SYNC_CAL_CONFIG, CALENDAR, CALENDAR_TYPE, Configuration.calendarSourceUri, read);
        if (restore)
        {
            calendarSourceConfig.setLastAnchor(last);
            calendarSourceConfig.setNextAnchor(next);
        }
        SyncClient.saveSourceConfig(SyncClient.SYNC_CAL_CONFIG, calendarSourceConfig);
        
        
        restore = false;
        if (noteSourceConfig != null)
        {
            last = noteSourceConfig.getLastAnchor();
            next = noteSourceConfig.getNextAnchor();
            restore = true;
        }
        noteSourceConfig     = SyncClient.initSourceConfig(SyncClient.SYNC_NOTE_CONFIG, NOTE, NOTE_TYPE, Configuration.noteSourceUri, read);
        if (restore)
        {
            noteSourceConfig.setLastAnchor(last);
            noteSourceConfig.setNextAnchor(next);
        }
        SyncClient.saveSourceConfig(SyncClient.SYNC_NOTE_CONFIG, noteSourceConfig);
    
        
        restore = false;
        if (taskSourceConfig != null)
        {
            last = taskSourceConfig.getLastAnchor();
            next = taskSourceConfig.getNextAnchor();
            restore = true;
        }
        taskSourceConfig     = SyncClient.initSourceConfig(SyncClient.SYNC_TASK_CONFIG, TASK, TASK_TYPE, Configuration.taskSourceUri, read);
        if (restore)
        {
            taskSourceConfig.setLastAnchor(last);
            taskSourceConfig.setNextAnchor(next);
        }
        SyncClient.saveSourceConfig(SyncClient.SYNC_TASK_CONFIG, taskSourceConfig);
    }
    
    /**
     * Initalize a specific syncsource config
     * 
     * @param storageKey
     * @param name
     * @param type
     * @param remoteUri
     * @param read              True will read from the Configuration.  False will create a new config
     * @return
     */
    private static SourceConfig initSourceConfig(final String storageKey, final String name, final String type, final String remoteUri, boolean read)
    {
        StaticDataHelper.info("Initialize SourceConfig ["+name+"]");
        SourceConfig ret = new SourceConfig(name, type, remoteUri);
        if (read)
        {
            try
            {
                byte[] byte_array = Configuration.getValue(storageKey,new byte[0]);
                if (byte_array!=null)
                {
                    DataInputStream temp = new DataInputStream(new ByteArrayInputStream(byte_array));
                    ret.deserialize(temp);
                    temp.close();
                }
            }
            catch(final Exception e)
            {
                StaticDataHelper.error("Exception while initialization (read) of SourceConfig ["+name+"]");
            }
        }
        return ret;
    }

    /**
     * Saves a syncsource config to storage
     * 
     * @param storageKey
     * @param config
     */
    private static void saveSourceConfig(final String storageKey, SourceConfig config)
    {
        StaticDataHelper.trace("Storing SourceConfig ["+config.getName()+"]");
        try
        {
            ByteArrayOutputStream buff = new ByteArrayOutputStream(512);
            DataOutputStream temp = new DataOutputStream(buff);
            config.serialize(temp);
            Configuration.append(storageKey,buff.toByteArray());
            temp.close();
        }
        catch(final Exception e)
        {
            StaticDataHelper.error("Exception while storing SourceConfig ["+config.getName()+"]");
        }
    }
   
    /**
     * Immediately disable autosync
     */ 
    public void disablePollingPIM()
    {
        if (pollerPIM != null && pollerPIM.isAlive())
        {
            pollerPIM.disable();
            pollerPIM = null;
        }
    }

    /**
     * Enable autosync.  This will reset the timer
     */
    public void enablePollingPIM()
    {
        disablePollingPIM();
        
        try
        {
            pollerPIM = new Poller(this, Configuration.pollingPIMInterval,Configuration.enablePollingPIM);
            pollerPIM.start();
        }
        catch (final Throwable t)
        {
            StaticDataHelper.error("Exception " + t.getMessage()+ " creating a new SyncClient object for the Poller: " + t.toString());
            return;
        }
    }

    /**
     * Helper function to build the DevConf
     * 
     * @return  A DeviceConfig
     */
    private DeviceConfig getDeviceConfig()
    {
        if (devconf != null)
        {
            return devconf;
        }
        devconf = new DeviceConfig();
        devconf.man = DeviceInfo.getManufacturerName();
        devconf.mod = DeviceInfo.getDeviceName();
        devconf.swv = VER;
        devconf.hwv = "" + DeviceInfo.getPlatformVersion();
        devconf.devID = "fbb-" + DeviceInfo.getDeviceId();
        devconf.maxMsgSize = 32*1024;
        return devconf;
    }

    
    /**
     * return error message
     * @param error
     * @param remoteDBType
     * @param sourceUri
     */
    private String getSyncError(String error, String remoteDBType, String sourceUri)
    {
        if (ERROR_NOT_AUTHORIZED.equals(error))
        {
            return StaticDataHelper.getLanguage("message_auth-failed");
        }
        else if (ERROR_NOT_FOUND.equals(error))
        {
            return StaticDataHelper.getLanguage("message_syncsource-not-found") + ": " + sourceUri;
        }
        else if (ERROR_SERVER.equals(error))
        {
            return StaticDataHelper.getLanguage("message_internal-server-error");
        }
        else if (error.indexOf(ERROR_REFRESH_REQUIRED) >= 0)
        {// 508
            for (int x = 1; x < SYNC_ALL; x *= 2)
            {
                BBPIMSyncSource ss = (BBPIMSyncSource) sources.get(new Integer(x));
                if (remoteDBType.equals(ss.getName()))
                {
                        ss.setLastAnchor(01);
                }
            }
            return StaticDataHelper.getLanguage("message_refresh");
        }
        
        else if (error != null)
        {
            StaticDataHelper.error(error);
            return StaticDataHelper.getLanguage("message_generic-error");
        }
        else
        {// error == null
            // if we return just 'error', the string "null" is printed on the
            // device's screen!"
            StaticDataHelper
                .error(error);
            return StaticDataHelper.getLanguage("message_generic-error");
        }
    }

    
    /**
     * Overridden run method of Runnable interface. A seperate thread is created
     * to synchronize data with sync server.
     * 
     * 'thread' is always the same Thread object in this class, representing the
     * runnable class SyncClient itself. It is created by choosing the
     * Synchronize menu item
     * 
     * @param void
     * @return void
     */
    public void run()
    {
        try
        {
            final Thread current = Thread.currentThread();
            if (this.thread == current)
            {
                synchronized (this)
                {
                    synchronize();
                }
            }
        }
        catch (final Exception ex)
        {
            StaticDataHelper.error("SyncClient.run(): Exception -> " + ex.toString());
            appScreen.getSyncScreen().displayFailure(
                        StaticDataHelper.getLanguage("sync_exception") + " " + ex);
        }
    }

    /**
     * Get the AppScreen
     */
    public AppScreen getAppScreen()
    {
        return appScreen;
    }

    /**
     * Initiate a refresh for some data types
     * 
     * @param syncMask  The mask of the types to sync
     */
    public void refresh(int syncMask, int direction)
    {
        doRefresh = true;
        refreshDirection = direction;
        synchronize(syncMask);  
    }
    
    /**
     * Invoked when the synchronize button in the menu of the home screen is
     * clicked. In this case a thread representing this runnable class
     * (SyncClient) is created and the run() method implemented by SyncClient is
     * fired.
     * 
     * @see run()
     */
    public void synchronize()
    {
        Log.setLogLevel(Configuration.logLevel);
        synchronize(SYNC_SUPPORTED);
    }

    
    /**
     * This method starts a synchronization on the specified sources, if they
     * are also configured by the user. It is used both for user and server
     * initialized sync
     * 
     * @param syncMask
     *        A bit mask of sources to sync
     */
    public void synchronize(final int syncMask)
    {
       
        if (Configuration.username == null || Configuration.username.length() == 0 || 
                        Configuration.password == null || Configuration.password.length() == 0)
        {
                showMessage(StaticDataHelper.getLanguage("message_login-required"));
                return;
        }
        
        if (BlackberryHelper.areConfigsEmpty() && !BlackberryHelper.workingConfigHasBeenFound()) {
            System.out.println("configs are empty, reinit!");
            BlackberryHelper.initConfigs();
            if (BlackberryHelper.shouldAskUserBeforeConnecting()) {
                Log.debug("setting connection handler to appscreen");
                BlackberryHelper.setConnectionHandler(appScreen);
            } else {
                Log.debug("setting connection handler to basic connection handler");
                BlackberryHelper.setConnectionHandler(new BasicConnectionHandler());
            }
        }
        
        
        if (SyncClient.checkCustom(SyncClient.CUSTOMS_UPDATE))
        {
            checkUpdate(true, false);
        }
        doCancel = false;
        StaticDataHelper.info("SyncClient.synchronize(int): Entering - 'inSynchronize' is " + inSynchronize);
        try
        {
            if (!Configuration.syncContact && !Configuration.syncCalendar
                    && !Configuration.syncTask && !Configuration.syncNote
                    && !Configuration.syncMail)
            {
                showMessage(
                    StaticDataHelper.getLanguage("message_nothing-to-sync")
                 );              
                return;
            }
            boolean signal = true;
            if (RadioInfo.getState() != RadioInfo.STATE_ON
                || RadioInfo.getSignalLevel() == RadioInfo.LEVEL_NO_COVERAGE)
            {
                signal = false;
            }
            if (!signal)
            {
                if (RadioInfo.getState() == RadioInfo.STATE_OFF)
                {
                    showMessage(
                        StaticDataHelper.getLanguage("message_radio-off")
                    );
                }
                else
                {
                    showMessage(
                        StaticDataHelper.getLanguage("message_no-signal")
                    );
                }
                return;
            }
            if (syncMask != 0)
            {
                checkDeletes(syncMask);
                if (doCancel)
                        return;
                
                showScreen(appScreen.getSyncScreen());
                
                final Thread syncThread = new Thread()
                {
                    public synchronized void run()
                    {
                        if (inSynchronize)
                        {
                            // Notify the user that another sync is in progress
                            showScreen(appScreen.getSyncScreen());
                            return;
                        }
                        inSynchronize = true;
                        
                        boolean wasRefresh = SyncClient.this.doRefresh;

                        try
                        {
                            StaticDataHelper.info("SyncClient.synchronize(int): Starting sync thread - 'inSynchronize' is " + inSynchronize);
                            SyncClient.this.syncThreadRun(syncMask);
                            compressionRetry = false;
                        }catch (CompressedSyncException e) {
                            if (!compressionRetry)
                            {
                            	compressionRetry = true;
                            }
                            else
                            {
                            	// Should never happen!
                            	compressionRetry = false;
                                StaticDataHelper.error("Sync failed with compression error after disabling compression");
                                appScreen.getSyncScreen().displayFailure(
                                        StaticDataHelper.getLanguage("sync_failed")
                                );
                            }
                        }catch (Exception e)
                        {
                            String message = e.getMessage();
                            compressionRetry = false;
                            if (message == null)
                            {
                                message = StaticDataHelper.getLanguage("message_no-error");
                            }
                            appScreen.getSyncScreen().displayFailure(
                                        StaticDataHelper.getLanguage("sync_failed") + " - "
                                        + message
                                );
                        }
                        finally
                        {
                            inSynchronize = false;
                            // Retry without compression.
                            // This must be in the finally block to avoid race conditions!
                            if (compressionRetry)
                            {
	                            // Only retry because of compression once
	                            StaticDataHelper.info("Sync failed because compression failed - Retrying");
	                            // Refresh direction is not cleared, so all we need to do is restore the flag
	                            SyncClient.this.doRefresh =  wasRefresh; 
	                            SyncClient.this.runUncompressedSync(syncMask);
                            }

                        }
                    }
                };
                syncThread.start();
            }// end if
        }
        catch (final Exception e)
        {
            appScreen.getSyncScreen().displayFailure(e.getMessage());
        }
    }

    
    /**
     * The main procedure for the sync thread
     * 
     * @param syncMask
     * @throws Exception
     */
    public void syncThreadRun(int syncMask) throws Exception
    {
        StaticDataHelper.getErrors().clearErrors();
        
        sc.syncUrl           =       (Configuration.syncUrl);
        sc.lastServerUrl     =       (Configuration.syncUrl);
        sc.userName          =       (Configuration.username);
        sc.password          =       (Configuration.password);
        sc.userAgent         =       ("Funambol BlackBerry Plug-in " + BBVersion.VERSION);
        sc.deviceConfig      =       (getDeviceConfig());
        //sc.compress = false;
        SyncManager manager = new SyncManager(sc);
        
        int syncMode = SyncML.ALERT_CODE_FAST;
        boolean showTCPAlert = false;
        String status;
        
        if (doRefresh)
        {
            if (refreshDirection == SyncClient.REFRESH_FROM_SERVER)
            {
                syncMode = SyncML.ALERT_CODE_REFRESH_FROM_SERVER;
            }
            else if (refreshDirection == SyncClient.REFRESH_TO_SERVER)
            {
                syncMode = SyncML.ALERT_CODE_REFRESH_FROM_CLIENT;
            }
            else
            {
                StaticDataHelper.error("Invalid refresh code");
                throw new Exception("Invalid refresh code");
            }
            doRefresh = false;
            status = StaticDataHelper.getLanguage("sync_recover-in-progress");
        }
        else
        {
            switch (Configuration.syncType){
                case StaticDataHelper.SYNC_TYPE_TWO_WAY:
                default:
                        syncMode = SyncML.ALERT_CODE_FAST;
                break;
                
                case StaticDataHelper.SYNC_TYPE_ONE_WAY_TO_CLIENT:
                        syncMode = SyncML.ALERT_CODE_ONE_WAY_FROM_SERVER;
                break;
                
                case StaticDataHelper.SYNC_TYPE_ONE_WAY_TO_SERVER:
                        syncMode = SyncML.ALERT_CODE_ONE_WAY_FROM_CLIENT;
                break;
            }
            
            status = StaticDataHelper.getLanguage("sync_in-progress");
        }
        
        if (!compressionRetry)
        {
            appScreen.getSyncScreen().displayStatus(
                    StaticDataHelper.getLanguage("sync_init")
                );
            appScreen.getSyncScreen().displayStatus(
                    SYNC_SUPPORTED, StaticDataHelper.getLanguage("sync_init")
                );
        }

        appScreen.getSyncScreen().setGlobalStatus("");

        if (!Configuration.syncContact)
                syncMask &= ~SYNC_CONTACTS;
        if (!Configuration.syncCalendar || !enableSyncCalendar)
                syncMask &= ~SYNC_CALENDAR;
        if (!Configuration.syncTask)
                syncMask &= ~SYNC_TASKS;
        if (!Configuration.syncNote)
                syncMask &= ~SYNC_NOTES;
        
        for (int x = 1; x <= SYNC_SUPPORTED; x*=2)
        {
             if ((syncMask & x) != 0)
             {
                appScreen.getSyncScreen().displayStatus(x,StaticDataHelper.getLanguage("sync_wait"));
                appScreen.getSyncScreen().setClientCount(x, 0, 0);
                appScreen.getSyncScreen().setServerCount(x, 0, 0);
             }
             else
             {
                appScreen.getSyncScreen().displayDisabled(x,StaticDataHelper.getLanguage("sync_disabled"));
                appScreen.getSyncScreen().setClientCount(x, 0, 0);
                appScreen.getSyncScreen().setServerCount(x, 0, 0);
             }
        }


        if (SyncClient.checkCustom(SyncClient.CUSTOMS_UPDATE))
        {
            if (!compressionRetry)
            {
                appScreen.getSyncScreen().displayStatus(
                        StaticDataHelper.getLanguage("sync_check-update")
                );

                checkUpdate(true, true);
        
                if (SyncClient.this.updateRequired){
                    SyncClient.showMessage(StaticDataHelper.getLanguage("message_update-required"));
                    SyncClient.getApplication().requestForeground();
                }
            }
        }

        appScreen.getSyncScreen().displayStatus(status);
        
        int numfail = 0;
        for (int thisSyncMask = 1; thisSyncMask <= SYNC_SUPPORTED; thisSyncMask*=2)
        {
            String type = null;
            String uri = null;
            SourceConfig ssc = null;
            String sscKey = null;
            boolean doSync = false;
            
            if ((syncMask & thisSyncMask) != 0)
            {
                switch (thisSyncMask)
                {
                    case SyncClient.SYNC_CONTACTS:
                            type = "contacts";
                            uri = Configuration.contactSourceUri;
                            ssc = contactSourceConfig;
                            sscKey = SyncClient.SYNC_CONTACT_CONFIG;
                            doSync = Configuration.syncContact;
                            break;
                            
                    case SyncClient.SYNC_CALENDAR:
                            type = "calendar";
                            uri = Configuration.calendarSourceUri;
                            ssc = calendarSourceConfig;
                            sscKey = SyncClient.SYNC_CAL_CONFIG;
                            if (enableSyncCalendar){
                                doSync = Configuration.syncCalendar;
                            }else{
                                doSync = false;
                            }
                            break;
                            
                    case SyncClient.SYNC_TASKS:
                            type = "tasks";
                            uri = Configuration.taskSourceUri;
                            ssc = taskSourceConfig;
                            sscKey = SyncClient.SYNC_TASK_CONFIG;
                            doSync = Configuration.syncTask;
                            break;
                            
                    case SyncClient.SYNC_NOTES:
                            type = "notes";
                            uri = Configuration.noteSourceUri;
                            ssc = noteSourceConfig;
                            sscKey = SyncClient.SYNC_NOTE_CONFIG;
                            doSync = Configuration.syncNote;
                            break;
                            
                    default:
                            doSync = false;
                            break;
                }
                if (doSync)
                {
                    if (doCancel)
                    {
                        appScreen.getSyncScreen().displayFailure(thisSyncMask,
                                StaticDataHelper.getLanguage("sync_cancelled")
                        );
                    }
                    else
                    {
                        appScreen.getSyncScreen().displayStatus(
                                    thisSyncMask,
                                    StaticDataHelper.getLanguage("sync_syncing"));
                        
                        BBPIMSyncSource source = (BBPIMSyncSource) sources.get(new Integer(thisSyncMask));
                        source.setConfig(ssc);
                        
                        try
                        {
                            manager.setFlagSendDevInf(); // TODO not always necessary
                            manager.sync(source, syncMode);
                            appScreen.getSyncScreen().setGlobalStatus("");

                            if (source.getErrorCount() != 0)
                            {
                                appScreen.getSyncScreen().displayError(thisSyncMask,
                                        StaticDataHelper.getLanguage("sync_complete-errors")
                                        );
                            }
                            else if (source.getWarningCode()!= null)
                            {
                                appScreen.getSyncScreen().displayWarning(thisSyncMask,
                                        StaticDataHelper.getLanguage("sync_complete-warnings")
                                        );
                            }
                            else
                            {
                                appScreen.getSyncScreen().displayStatus(thisSyncMask,
                                        StaticDataHelper.getLanguage("sync_complete")
                                        );
                            }
                            
                            SyncClient.saveSourceConfig(sscKey,ssc);
                        } catch (final SyncException e) {
                            numfail++;
                            source.restoreTrackingData();
                            String message = "";
                            String message2 = "";
                            try {
                                message = getSyncError(e.getMessage(), type, uri);
                                message2 = message;
                            } catch (Exception e2) {
                                message = StaticDataHelper.getLanguage("message_no-error");
                                message2 = "Exception type: " + e.getClass().getName();
                            }
                            
                            boolean compressError = (e instanceof CompressedSyncException);
                            
                            if (compressError && !SyncClient.this.compressionRetry)
                            {
                            	message2 += " - Retrying without compression";
                            	StaticDataHelper.info(message2);
                            }
                            
                            if (!compressError || (compressError && SyncClient.this.compressionRetry))
                            {
                                appScreen.getSyncScreen().displayFailure(thisSyncMask,
                                        StaticDataHelper.getLanguage("sync_failed") + " - " + message
                                );
                                StaticDataHelper.error(message2);
                            }
                            
                            if (compressError)
                            {
                                throw new CompressedSyncException(e.getMessage());
                            }
                            
                            int code = e.getCode();
                            if (code == SyncException.READ_SERVER_RESPONSE_ERROR || 
                                code == SyncException.WRITE_SERVER_REQUEST_ERROR || 
                                code == SyncException.CONN_NOT_FOUND
                                )
                            {
                                displayConnectivityError = true;
                            }
                            
                             
                              if( code == SyncException.CONNECTION_BLOCKED_BY_USER) {
                                   showTCPAlert = true ;
                               }
                             
                            
                            
                        }
                        appScreen.getSyncScreen().setGlobalStatus("");
                        source = null;
                    }
                }
            }
        }

        doCancel = false;
        
        // -- END OF A SYNCHRONIZATION
        if (numfail > 0 )
        {
            appScreen.getSyncScreen().displayFailure(StaticDataHelper.getLanguage("sync_complete-errors"));
            showScreen(appScreen.getSyncScreen());
            Application.getApplication().requestForeground();
            
            //TODO: fix this, it's horrible to use a flag like this.
            
            if (showTCPAlert) {
                Log.debug("showing tcp settings alert!");
                SyncClient.showMessage(StaticDataHelper.getLanguage("message_enter-TCP-settings"));
                           
            }
        }
        else
        {
            appScreen.getSyncScreen().displaySuccess(StaticDataHelper.getLanguage("sync_complete"));
        }
        
        if (displayConnectivityError)
        {
            if (syncType.equals(MANUAL_SYNC)){
                Log.error("Manual sync: Connection attempts failed.");
                SyncClient.showMessage(StaticDataHelper.getLanguage("message_connectivity"));
                displayConnectivityError = false;
            }else if (syncType.equals(SCHEDULED_SYNC)){
                
                if(pol!=null){
                    pol.disable();
                }
                
                if(scheduledAttempt < 3){
                    scheduledAttempt++;
                    Log.error("Scheduled sync: Connection attempt failed. " +
                              "Try again between "+NEW_POL_TIME+" minutes");

                    pol = new Poller(this, NEW_POL_TIME, true);
                    pol.start();
                }else{
                    pol.disable();
                    pol = null;
                    scheduledAttempt = 0;
                    SyncClient.showMessage(StaticDataHelper.getLanguage("message_connectivity"));
                    displayConnectivityError = false;
                }
            }
            
            
        }
        
        
    }// end run()
   
    
    /**
     * Cancel the current sync
     */
    public void cancelSync()
    {
        doCancel = true;
        if (inSync())
        {
            
            appScreen.getSyncScreen().displayStatus(StaticDataHelper.getLanguage("sync_cancelling"));
            
        }
    }
    
    /**
     * Check if the current sync should be cancelled
     * 
     * @return
     */
    public boolean isCancelled()
    {
        return doCancel;
    }
    
    /**
     * Requests minimization of the program
     **/
    public void minimize()
    {
        getApplication().requestBackground();
    }

    /**
     * Returns true if there is current a sync in progress
     **/
    public boolean inSync()
    {
        return inSynchronize;
    }
    
    
    /**
     * Checks update.xml (location in language.xml) for updates to the plugin.
     * 
     * @param prompt    If true, the user is prompted if there is an update
     * @param wait              If true, the procedure waits for the process to complete.
     *                                  If false, a new thread is spawned.
     */
    public void checkUpdate(final boolean prompt, final boolean wait)
    {
        final String UPDATE_ROOT_TAG = "sync";
        final String BB_UPDATE_TAG = "blackberry";
        final String VERSION_TAG = "version";
        final String REQUIRED_TAG = "required";
        final String URL_TAG = "url";
        final String OTA_TAG = "ota";
        final String VER_SEP = ".";

        if (RadioInfo.getState() == RadioInfo.STATE_OFF ||
            RadioInfo.getSignalLevel() == RadioInfo.LEVEL_NO_COVERAGE)
        {
            return;
        }
                
        Thread check = new Thread()
        {
            public void run()
            {
                // TODO Strip all trailing directories from sync server uri
                // and append update.xml
                String url = StaticDataHelper.getLanguage("update_xml");
    
                HttpConnection c = null;
                InputStream is = null;
                int rc;
                
                try {
                    StaticDataHelper.info(url);
                    c = (HttpConnection)Connector.open(url);
                    // Getting the response code will open the connection,
                    // send the request, and read the HTTP response headers.
                    // The headers are stored until requested.
        
                    rc = c.getResponseCode();
                    if (rc != HttpConnection.HTTP_OK) 
                    {
                        return;
                    }
                    is = c.openInputStream();
        
                    SimpleXMLParser parser = new SimpleXMLParser();
                    parser.addHashtableValueTag(BB_UPDATE_TAG);
                    
                    Hashtable output = parser.parseData(UPDATE_ROOT_TAG,is);
                    Hashtable bb = (Hashtable)output.get(BB_UPDATE_TAG);
                    if (bb!=null)
                    {
                        String version = SyncClient.VER;
                        String fversion = (String)bb.get(VERSION_TAG);
        
                        String required = (String)bb.get(VERSION_TAG);
                        
                        int vpos = 0;
                        int vfpos = 0;
                        do
                        {
                            vpos = version.indexOf(VER_SEP,0);
                            vfpos = fversion.indexOf(VER_SEP,0);
                            if (vpos < 0 && version.length()>0)
                            {
                                vpos = version.length();
                            }
                            if (vfpos < 0 && fversion.length()>0)
                            {
                                vfpos = fversion.length();
                            }
        
                            if (vpos > 0 && vfpos > 0)
                            {
                                int val = Integer.parseInt(version.substring(0,vpos));
                                int fval = Integer.parseInt(fversion.substring(0,vfpos));
                                if (val < fval)
                                {
                                    SyncClient.this.updateFlag = true;
                                    break;
                                }
                                else if (val > fval)
                                {
                                        SyncClient.this.updateFlag = false;
                                    break;
                                }
                                if (vpos<version.length())
                                {
                                    version = version.substring(vpos+1);
                                }
                                else
                                {
                                    vpos = -1;
                                }
                                if(vfpos<fversion.length())
                                {
                                    fversion = fversion.substring(vfpos+1);
                                }
                                else
                                {
                                    vfpos = -1;
                                }
                            }
                        }
                        while(vpos > 0 && vfpos > 0);
                        
                        if (SyncClient.this.updateFlag)
                        {
                            SyncClient.this.updateOTA = (String)bb.get(OTA_TAG);
                            SyncClient.this.updateURL = (String)bb.get(URL_TAG);
                            SyncClient.this.updateRequired = !((String)bb.get(REQUIRED_TAG)).equals("0");
                        }
                    }
                    else
                    {
                        StaticDataHelper.info("BlackBerry info set is null");
                    }
                    output = null;
                    bb = null;
                } 
                catch (ClassCastException cce) 
                {
                    StaticDataHelper.info(cce.toString());
                } 
                catch (Throwable t) 
                {
                    StaticDataHelper.info(t.toString());
                } 
                finally 
                {
                    if (is != null)
                        try { is.close(); } catch (Exception e) {}
                    if (c != null)
                        try { c.close();  } catch (Exception e) {}
                }
        
                if (SyncClient.this.updateFlag && prompt)
                {
                    showMessage(StaticDataHelper.getLanguage("message_update-available"));
                    SyncClient.getApplication().requestForeground();
                }
            }
        };
        
        check.start();
        if (wait)
        {
            try 
            {
                check.join();
            }
            catch (Throwable e)
            {
                    
            }
        }
    }
    
    /**
     * Returns the current status of the update flag 
     * @return  True if there is an update available.
     */
    public boolean isUpdate()
    {
        return updateFlag;
    }
    
    /**
     * Called when user selected Download Update from the menu
     */
    public void doUpdate()
    {
        checkUpdate(false, true);
        if (updateFlag)
        {
            boolean result = askYesNoQuestion(
                            StaticDataHelper.getLanguage("dialog_update"), true);
            if (result)
            {
                Browser.getDefaultSession().displayPage(updateOTA);
            }
            else
            {
                showMessage(StaticDataHelper.getLanguage("message_update-location") + ":\n" + updateURL);
            }
        }
    }

    /**
     * Called to warn the user of a large number of deletes
     * @param mask      Bit mask of affected syncsource
     */
    public void warnDelete(int mask)
    {
        Application.getApplication().requestForeground();
        
        StaticDataHelper.info("Prompting user for delete confirmation");
        boolean result = askYesNoQuestion(
                        StaticDataHelper.getLanguage("dialog_delete"), false);
        if (!result)
        {
            StaticDataHelper.info("User opted to cancel sync for calendar");
             if (!Configuration.syncContact && !Configuration.syncTask && !Configuration.syncNote
                    && !Configuration.syncMail){
                  cancelSync();
             }else{
                enableSyncCalendar = false;
             }
        }
        else
        {
            StaticDataHelper.info("Continuing with sync");
            enableSyncCalendar = true;
        }
    }
    
    /**
     * Check to see if the user should be warned, and if so, warn them.
     * 
     * @param mask
     */
    private void checkDeletes(int mask)
    {
        if (!Configuration.syncContact)
                mask &= ~SYNC_CONTACTS;
        if (!Configuration.syncCalendar)
                mask &= ~SYNC_CALENDAR;
        if (!Configuration.syncTask)
                mask &= ~SYNC_TASKS;
        if (!Configuration.syncNote)
                mask &= ~SYNC_NOTES;
        
        for (int x = 1; x <= SYNC_SUPPORTED; x*=2)
        {
            BBPIMSyncSource source = (BBPIMSyncSource)sources.get(new Integer(x));
            if (!doRefresh && source != null && source.shouldWarn())
                warnDelete(x);
        }
    }
    
    /**
     * This is the entry point for SyncClient Application. The thread that calls
     * this method (typically the main thread in the application) becomes the
     * event-dispatching thread, which will execute all drawing and
     * event-handling code
     * 
     * @param s[]
     *        An array of strings passed on the command line as parameters
     *        for the <code>java</code> interpreter
     */
    public static void main(String args[]) throws Exception
    {
        if (SyncClient.checkCustom(SyncClient.CUSTOMS_ROLLOVER))
        {
            try
            {
                net.rim.blackberry.api.homescreen.HomeScreen.updateIcon(
                            Bitmap.getBitmapResource("icon.png")
                            );
                net.rim.blackberry.api.homescreen.HomeScreen.setRolloverIcon(
                            Bitmap.getBitmapResource("icon_over.png")
                            );   
            }
            catch(Exception e)
            {}
        }

        final SyncClient client = new SyncClient();
        if (client.loadVersion())
        {
            client.enterEventDispatcher();
        }
    }
    
    private void runUncompressedSync(int syncMask) {
        // Disable only-in-memory property for compression
        sc.compress = false;

        // Not really necessary, but it makes the code make sense
        compressionRetry = true;

        synchronize(syncMask);
    }

    public static boolean checkCustom(int flag)
    {
        if (!SyncClient.CUSTOMS_ENABLED)
            return false;
        if ((SyncClient.CUSTOMS & flag) != 0)
            return true;
        return false;
    }
}
