/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.helper;

import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import java.util.TimeZone;
import java.util.Vector;

import javax.microedition.pim.PIMItem;
import javax.microedition.pim.PIMList;
import javax.microedition.pim.PIMException;

import net.rim.device.api.util.StringUtilities;

import com.funambol.util.SimpleXMLParser;
import com.funambol.util.StaticDataHelper;
import com.funambol.util.XmlUtil;

public abstract class PIMItemHelper 
{
    public static final String      EMPTY_STRING        = "";
    public static final long        SECOND_FACTOR       = 1000;
    public static final long        MINUTE_FACTOR       = 60*PIMItemHelper.SECOND_FACTOR;
    public static final long        HOUR_FACTOR         = 60*PIMItemHelper.MINUTE_FACTOR;
    public static final long        DAY_FACTOR          = 24*PIMItemHelper.HOUR_FACTOR;
    public static final long        MONTH_FACTOR        = 30*PIMItemHelper.DAY_FACTOR;
    
    protected int limitPast;
    protected int limitFuture;
    protected boolean deleteForUpdate;
    
    
    public PIMItemHelper(int pastlimit, int futurelimit)
    {
        this.limitPast = pastlimit;
        this.limitFuture = futurelimit;
        this.deleteForUpdate = false;
    }
    
    public void setDeleteForUpdate(boolean flag)
    {
        this.deleteForUpdate = flag;
    }
    
    public boolean getDeleteForUpdate()
    {
        return this.deleteForUpdate;
    }
    
    /**
     * This function will try to convert a PIMItem to appropriate SIF format.
     * @param pimList The PIMList will be used for checking if a Property is supported
     * @param item The PIMItem will be converted
     * @return An XML - SIF string or null.
     */
    public abstract String toString(PIMList pimlist,PIMItem item) throws PIMException;
    
    /**
     * This function will try to get the UID string of provided PIMItem.
     * @param item The PIMItem
     * @return A string which is the UID of the provided PIMItem or null
     */
    public abstract String getUID(PIMItem item) throws PIMException;
    
    /**
     * The function will try to create a new-blank PIMItem
     * @param pimlist The PIMList that the new PIMItem will be created from.
     * @return A new-blank PIMItem or null
     */
    public abstract PIMItem createPIMItem(PIMList pimlist) throws PIMException;
    
    /**
     * This function will search through the provided PIMList to find (and delete) the PIMItem 
     * that has match itemKey (UID)
     * @param pimlist The PIMList that will be searched through
     * @param itemKey The UID of the PIMItem that we want.
     * @param deleteFlag The boolean flag to tell if the found PIMItem should be deleted or not.
     * @return The found PIMItem or null.
     */
    public abstract PIMItem findPIMItem(PIMList pimlist, String itemKey, boolean deleteFlag) throws PIMException;
    public abstract Hashtable findPIMItems(PIMList pimlist, Vector itemKeys, boolean deleteFlag) throws PIMException;
    
    /**
     * This function will parse a XML - SIF string and set the values to provided PIMItem.
     * @pram pimlist An instance of PIMList
     * @param item A PIMItem 
     * @param pimString The XML-SIF format string
     */
    public abstract void setPIMProperties(PIMList pimlist, PIMItem item,String pimString) throws PIMException;
    
    public abstract void deleteAllPIMItems(PIMList pimlist) throws PIMException;
    
    public abstract boolean isInSyncRange(PIMList pimlist,PIMItem _item) throws PIMException;
    
    /**
     * This function will clear off all properties of a PIMItem with a value.
     * @param _item The PIMItem that will have all properties cleared off.
     */
    public abstract void clearPIMItem(PIMItem _item) throws PIMException;
    
    public PIMItem findPIMItem(PIMList pimlist, String itemKey) throws PIMException
    {
        return this.findPIMItem(pimlist,itemKey,false);
    }
    
    public Hashtable findPIMItems(PIMList pimlist, Vector itemKeys) throws PIMException
    {
        return this.findPIMItems(pimlist,itemKeys,false);
    }
    
    public PIMItem removePIMItem(PIMList pimlist, String itemKey) throws PIMException
    {
        return this.findPIMItem(pimlist,itemKey,true);
    }
    
    public static void clearPIMItem(PIMItem _item, int _UID, int[] fieldIDs) throws PIMException
    {
        int oldvalues[] = _item.getFields();
        for(int i=0;i<oldvalues.length;i++)
        {
            if (_UID != oldvalues[i])
            {
                for (int t=0;t<fieldIDs.length;t++)
                {
                    if (oldvalues[i] == fieldIDs[t])
                    {
                        int count = _item.countValues(oldvalues[i]);
                        StaticDataHelper.debug("Field ID ["+oldvalues[i]+"] has ["+count+"] values");
                        for (int j=count-1;0<=j;j--)
                        {
                            _item.removeValue(oldvalues[i],j);
                            StaticDataHelper.debug("Removing value ["+j+"] of Field ID ["+oldvalues[i]+"]");
                        }
                        count = _item.countValues(oldvalues[i]);
                        StaticDataHelper.debug("After deletion Field ID ["+oldvalues[i]+"] has ["+count+"] values");                
                        break;
                    }
                }
            }
        }
        String oldCategories[] = _item.getCategories();
        for(int i=0;i<oldCategories.length;i++)
        {
            _item.removeFromCategory(oldCategories[i]);
            StaticDataHelper.debug("Remove old category ["+oldCategories[i]+"]");
        }
        oldvalues = null;
    }
    
    /**
     * This method checks if the field and attibute are supported by the PIMList
     * @param int field
     * @param int attribute
     * @return boolean:returns true if both the field and attribute associated with it r supported
     */
    public static boolean isSupportedAttributedField(PIMList list, int field, int attribute)
    {
        return PIMItemHelper.isSupportedField(list, field) && list.isSupportedAttribute(field, attribute);
    }
    
    /**
     * Checks if the type of field is supported by the PIMList
     * @param int type
     * @return boolean returns true if the contact list supports this field
     */
    public static boolean isSupportedField(PIMList list, int type)
    {
        return list.isSupportedField(type);
    }
    
    public static Vector getVectorValue(Hashtable eventMap, String key)
    {
        final Object val = eventMap.get(key);
        final Vector value = val == null ? new Vector(0) : (Vector)val;
        return value;
    }
    
    public static String getValue(Hashtable eventMap, String key, boolean unescape)
    {
        final Object val = eventMap.get(key);
        if (val==null)
        {
            return PIMItemHelper.EMPTY_STRING;
        }
        return unescape ? XmlUtil.unescapeXml(val.toString().trim()) : val.toString().trim();
    }
    
    /**
     * This function will add Category data to provided PIMItem.
     * 
     * @param pimMap The hashtable that was created from parsing the SIF data.
     * @param list This is the PIMList that the PIMItem belongs to.
     * @param _item The PIMItem that will have Category data added to it.
     * @param CategoryTag The string which is SIF tag for Categories.
     */
    public static void addCategories(Hashtable pimMap, PIMList list, PIMItem _item, String CategoryTag) throws PIMException
    {
        String field = PIMItemHelper.getValue(pimMap, CategoryTag, false);// Categories
        int max = _item.maxCategories();
        max = max < 0 ? Integer.MAX_VALUE : max;
        final String[] categories = StringUtilities.stringToWords(field);
        if (categories.length > max)
        {
            StaticDataHelper.info("Some categories being dropped (max = " + max + ", length = " + categories.length + ")");
        }
        for (int j = 0; j < (max < categories.length ? max : categories.length); j++)
        {
            if (list.isCategory(categories[j]))
            {
                try {
                    _item.addToCategory(categories[j]);
                }
                catch (PIMException e) {
                    StaticDataHelper.info("Exception while adding category ["+categories[j]+"]");
                }
            }
        }
    }
    
    /**
     * This function will try to extract the Categories out of a PIMItem.
     * 
     * @param _item The PIMItem will have categories extracted from.
     * @return An empty string or a tring in following format "<category>; <category>"
     */
    public static String extractCategories(PIMItem _item)
    {
        final String[] categories = _item.getCategories();
        final StringBuffer categoriesBuffer = new StringBuffer();
        String categoriesContents = null;
        if (categories.length > 0)
        {
            for (int k = 0; k < categories.length; k++)
            {
                categoriesBuffer.append(categories[k]);
                if (categories.length > 1 && k < categories.length - 1)
                {
                    categoriesBuffer.append("; ");// the separator (semicolon + blank)
                }
            }
            // e.g. Business; Marketing
            categoriesContents = categoriesBuffer.toString();
        }
        else
        {
            // no categories at all!
            categoriesContents = "";
        }
        return categoriesContents;
    }

    /**
     * This function parses a SIF format String to a Hashtable of fields.
     * @param contactString Information of a PIM in SIF format
     * @return A Hashtable with all fields from the above string
     */
    private static SimpleXMLParser _Parser; 
    public static Hashtable buildMapFromXML(String rootTag, String sifString, boolean resetparser
                                            , Vector multiValueTags, Vector hashtableValueTags, Vector multiHashtableBalueTags)
    {
        //StaticDataHelper.info("buildMapFromXML - starting - " + Calendar.getInstance().getTime().toString()); 
        try
        {
            if (PIMItemHelper._Parser == null)
            {
                PIMItemHelper._Parser = new SimpleXMLParser();
            }
            if (resetparser)
            {
                PIMItemHelper._Parser.reset();
            }
            // Multi simple value tags
            if (multiValueTags!=null)
            {
                for(int i=multiValueTags.size()-1;0<=i;i--)
                {
                    PIMItemHelper._Parser.addMultiValueTag((String)multiValueTags.elementAt(i));
                }
            }
            // Hashtable value tags
            if (hashtableValueTags!=null)
            {
                for(int i=hashtableValueTags.size()-1;0<=i;i--)
                {
                    PIMItemHelper._Parser.addHashtableValueTag((String)hashtableValueTags.elementAt(i));
                }
            }
            // multi Hashtable value tags
            if (multiHashtableBalueTags!=null)
            {
                for(int i=multiHashtableBalueTags.size()-1;0<=i;i--)
                {
                    PIMItemHelper._Parser.addMultiHashtableValueTag((String)multiHashtableBalueTags.elementAt(i));
                }
            }
            
            //StaticDataHelper.info("buildMapFromXML - ending - " + Calendar.getInstance().getTime().toString());

            return PIMItemHelper._Parser.parseData(rootTag,sifString);
        }
        catch(Throwable e)
        {
            StaticDataHelper.error("Exception while parsing SIF-XML data ["+e.toString()+"]");
            StaticDataHelper.error("SIF-XML data ["+sifString+"]");
        }
        return null;
    }
    
    // If the year is prior to 1970, set it to 1970.
    public static long adjustDateToEpoch(long date)
    {
        final Calendar cal = Calendar.getInstance();
        cal.setTime(new Date(date));
        if (cal.get(Calendar.YEAR) < 1970)
        {
            cal.set(Calendar.YEAR, 1970);
        }
        return cal.getTime().getTime();
    }
    
    /**
     * Get time (a {@code long} value that holds the number of milliseconds
     * since midnight GMT, January 1, 1970) from date in "yyyy-MM-dd" String
     * format
     * @param field Date in "yyyy-MM-dd" String format
     * @return time at 00:00:00 from date
     */
    public static Calendar parseDate(String field)
    {
        int day = 0;
        int month = 0;
        int year = 0;
        
        Calendar date = null;
        
        if (field.length() == 10)
        {
                year = Integer.parseInt(field.substring(0, 4));
                month = Integer.parseInt(field.substring(5, 7));
                day = Integer.parseInt(field.substring(8, 10));
        }
        else
        {
                year = Integer.parseInt(field.substring(0, 4));
                month = Integer.parseInt(field.substring(4, 6));
                day = Integer.parseInt(field.substring(6, 8));
        }
        
        //date = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
        date = Calendar.getInstance(TimeZone.getDefault());
        
        date.set(Calendar.DAY_OF_MONTH, day);
        date.set(Calendar.MONTH, month - 1);
        year = year > 1970 ? year : 1970;
        date.set(Calendar.YEAR, year);
        date.set(Calendar.HOUR_OF_DAY, 0);
        date.set(Calendar.MINUTE, 0);
        date.set(Calendar.SECOND, 0);
         
        return date;
    }
    
    /**
     * Get time from date in "yyyyMMddTHHmmssZ" or "yyyyMMdd" or "yyyyMMddTHHmmss" String format
     * @param field date in "yyyyMMddTHHmmssZ" String format
     * @return A 'long' time in milliseconds from the Epoch
     */
    public static Calendar parseDateTime(String field)
    {
        int day = 0;
        int month = 0;
        int year = 0;
        int hour = 0;
        int minute = 0;
        int second = 0;
                
        Calendar date = null;
        
        if (field.length() <= 10)
                return PIMItemHelper.parseDate(field);

        year = Integer.parseInt(field.substring(0, 4));
        year = year > 1970 ? year : 1970;
        month = Integer.parseInt(field.substring(4, 6));
        day = Integer.parseInt(field.substring(6, 8));
        if (field.charAt(8)=='T')
        {
            hour = Integer.parseInt(field.substring(9, 11));
            minute = Integer.parseInt(field.substring(11, 13));
            second = Integer.parseInt(field.substring(13, 15));
        }
                
        if (field.charAt(field.length()-1) == 'Z')
        {
                date = Calendar.getInstance(TimeZone.getTimeZone("GMT")); //GMT
        }
        else
        {
                date = Calendar.getInstance(TimeZone.getDefault());
        }
        date.set(Calendar.DAY_OF_MONTH, day);
        date.set(Calendar.MONTH, month - 1);
        date.set(Calendar.YEAR, year);
        date.set(Calendar.HOUR_OF_DAY, hour);
        date.set(Calendar.MINUTE, minute);
        date.set(Calendar.SECOND, second);
        return date;
    }
    
    public static Calendar getStartOfDay(Calendar date)
    {
        date.set(Calendar.HOUR_OF_DAY, 00);
        date.set(Calendar.MINUTE, 00);
        date.set(Calendar.SECOND, 00);// 00 from 59 to 00 bugfix
        return date;
    }
    
    public static Calendar getEndOfDay(Calendar date)
    {
        date.set(Calendar.HOUR_OF_DAY, 23);// 00
        date.set(Calendar.MINUTE, 59);// 00
        date.set(Calendar.SECOND, 00);// 00 from 59 to 00 bugfix
        return date;
    }
    
    public static long getTimestamp(Calendar date)
    {
        return date.getTime().getTime();
    }
    
    public static long convertLocalToGMT(long localtime)
    {
        long GMTtime = localtime - PIMItemHelper.getTimeZoneOffSet(new Date(localtime));
        return GMTtime;
    }
    
    /**
     * This function was build to fix a BUG in get/set DateTime in BlackBerryToDo 4.2.1.
     * This function might return a wrong value round the DayLightSaving time stamp.
     * Using this is just a work around for the BUG in BlackBerry API 4.2.1
     * @param gmttime <description>
     * @return <description>
     */
    public static long convertGMTToLocal(long gmttime)
    {
        long Localtime = gmttime + PIMItemHelper.getTimeZoneOffSet(new Date(gmttime));
        return Localtime;
    }
    
    /**
     * @param date a local-date to read offset
     * @return timezone offset
     */
    public static long getTimeZoneOffSet(Date date)
    {
        long offset = 0;
        TimeZone zn = TimeZone.getDefault();
        Calendar local = Calendar.getInstance();
        local.setTime(date);
        /*
        * the offset to add to GMT to get local time, modified in case of
        * daylight savings
        */
        int time = (int)(local.get(Calendar.HOUR_OF_DAY)*PIMItemHelper.HOUR_FACTOR + local.get(Calendar.MINUTE)*PIMItemHelper.MINUTE_FACTOR  + local.get(Calendar.SECOND)*PIMItemHelper.SECOND_FACTOR);
        offset = zn.getOffset(1 // era AD
                            ,local.get(Calendar.YEAR)
                            ,local.get(Calendar.MONTH)
                            ,local.get(Calendar.DAY_OF_MONTH)
                            ,local.get(Calendar.DAY_OF_WEEK)
                            ,time);
        return offset;
    }
    
    public static long getDefaultTimeZoneOffset()
    {
        long offset = 0;
        TimeZone zn = TimeZone.getDefault();
        Calendar local = Calendar.getInstance();
        local.setTime(new Date(System.currentTimeMillis()));
        /*
        * the offset to add to GMT to get local time, modified in case of
        * daylight savings
        */
        int time = (int)(local.get(Calendar.HOUR_OF_DAY)*PIMItemHelper.HOUR_FACTOR + local.get(Calendar.MINUTE)*PIMItemHelper.MINUTE_FACTOR  + local.get(Calendar.SECOND)*PIMItemHelper.SECOND_FACTOR);
        offset = zn.getOffset(1 // era AD
                            ,local.get(Calendar.YEAR)
                            ,local.get(Calendar.MONTH)
                            ,local.get(Calendar.DAY_OF_MONTH)
                            ,local.get(Calendar.DAY_OF_WEEK)
                            ,time);
        return offset;
    }
    
    public static Calendar getFieldDate(PIMList list, PIMItem item, int field, boolean endday, boolean allday, boolean extraToGMT)
    {
        long value = 0;
        long offset = 0;
        long oneDay = (allday && endday) ? (PIMItemHelper.DAY_FACTOR) : 0;
                
        if (list.isSupportedField(field))
        {
            try
            {
                value = item.getDate(field, 0);
                //StaticDataHelper.debug("PIMItemHelper.getFieldDate(): Retrieving from device ["+new Date(value).toString()+"] timestamp ["+value+"]");

                //final Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
                final Calendar cal = Calendar.getInstance();
                
                value = PIMItemHelper.convertLocalToGMT(value);
                //StaticDataHelper.debug("PIMItemHelper.getFieldDate(): After adjust TimeZone offset [" + new Date(value).toString()+"]");
                if (extraToGMT)
                {
                    value = PIMItemHelper.convertLocalToGMT(value);
                    //StaticDataHelper.debug("PIMItemHelper.getFieldDate(): After extra adjust TimeZone offset [" + new Date(value).toString()+"]");
                }
                value -= oneDay;
                cal.setTime(new Date(value));
                //StaticDataHelper.debug("PIMItemHelper.getFieldDate(): After adjust 1 day (if needed) [" + new Date(value).toString()+"]");
                return cal;
            }
            catch (final Exception e)
            {
                StaticDataHelper.info("PIMItemHelper.getFieldDate(): unknown exception during retrieving and converting Date value ["+ e.getMessage() +"]");
            }
        }
        return null;
    }
    
    public static String formatAllDay(Calendar allday)
    {
        String ret = "";
        int tt =0;
        ret += allday.get(Calendar.YEAR) + "-";
        
        tt = allday.get(Calendar.MONTH)+1;
        ret += (tt<10 ? "0"+tt : ""+tt )+ "-";
        
        tt = allday.get(Calendar.DAY_OF_MONTH);
        ret += tt<10 ? "0"+tt : ""+tt;
        return ret;
    }
    
    public static String formatDateTimeUTC(Calendar datetime)
    {
        String ret = "";
        int tt = 0;
        ret += datetime.get(Calendar.YEAR);
        
        tt = datetime.get(Calendar.MONTH)+1;
        ret += tt<10 ? "0"+tt : ""+tt;
        
        tt = datetime.get(Calendar.DAY_OF_MONTH);
        ret += (tt<10 ? "0"+tt : ""+tt) + "T";
        
        tt = datetime.get(Calendar.HOUR_OF_DAY);
        ret += tt<10 ? "0"+tt : ""+tt;
        
        tt = datetime.get(Calendar.MINUTE);
        ret += tt<10 ? "0"+tt : ""+tt;
        
        tt = datetime.get(Calendar.SECOND);
        ret += (tt<10 ? "0"+tt : ""+tt) + "Z";
        return ret;
    }

    /**
     * This method appends the event information String Type supported by
     * blackberry to a string buffer.
     * @param StringBuffer: StringBuffer containing event information.
     * @param item: PIMItem
     * @param int: type of field
     * @param String: Name of field
     */
    public static boolean appendXMLFieldBoolean(PIMList list, StringBuffer pimBuffer, PIMItem item, int field, String tag)
    {
        boolean value = false;
        if (list.isSupportedField(field))
        {
            try
            {
                value = item.getBoolean(field, 0);
            }
            catch (final Exception e)
            {}
            if (value)
            {
                XmlUtil.addElementNewLine(pimBuffer, tag, "1",false);
            }
            else
            {
                XmlUtil.addElementNewLine(pimBuffer, tag, "0",false);
            }
        }
        return value;
    }
    
    
    public static String handlePIMException(PIMException e)
    {
        String error = "";
        StaticDataHelper.debug("PIMItemHelper: Handling PIM Exception");
        if (e.getMessage() != null)
        {
                StaticDataHelper.info("PIMException message: " + e.getMessage());
        }
        switch (e.getReason())
        {
                case PIMException.FEATURE_NOT_SUPPORTED:
                                error = "feature not supported";
                        break;
                case PIMException.GENERAL_ERROR:
                                error = "general error";
                        break;
                case PIMException.LIST_CLOSED:
                                error = "list closed";
                        break;
                case PIMException.LIST_NOT_ACCESSIBLE:
                                error = "list not accessible";
                        break;
                case PIMException.MAX_CATEGORIES_EXCEEDED:
                                error = "max categories exceeded";
                        break;
                case PIMException.UNSUPPORTED_VERSION:
                                error = "unsupported version";
                        break;
                case PIMException.UPDATE_ERROR:
                                error = "update error";
                        break;
                default:
                                error = "unknown error";
                        break;
        }
                StaticDataHelper.info("PIMException reason: " + error);
                return error;
    }
    
} 
