/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.helper;

import java.util.Calendar;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.TimeZone;
import java.util.Vector;

import javax.microedition.pim.PIMItem;
import javax.microedition.pim.PIMList;
import javax.microedition.pim.PIMException;
import javax.microedition.pim.Event;
import javax.microedition.pim.EventList;
import javax.microedition.pim.RepeatRule;

import net.rim.blackberry.api.pdap.BlackBerryEvent;
import net.rim.device.api.i18n.SimpleDateFormat;

import com.funambol.blackberry.helper.SIFRepeatRuleHelper;
import com.funambol.util.StaticDataHelper;
import com.funambol.util.XmlUtil;

public class SIFEventHelper extends PIMItemHelper
{
        private static final String             DATETIME_FORMAT_ALLDAY                  = "yyyy-MM-dd";
        private static final String             DATETIME_FORMAT_UTC                     = "yyyyMMdd'T'HHmmss'Z'";
    
        private static final String             ALARM                                   = "ReminderMinutesBeforeStart";
        private static final String             ALARM_ENABLE                            = "ReminderSet";
        private static final String             ALLDAY                                  = "AllDayEvent";
        private static final String             BUSY_STATUS                             = "BusyStatus";                     // bugfix
        private static final String             END                                     = "End";
        private static final String             IS_RECURRING                            = "IsRecurring";                    // enhancement
        private static final String             LOCATION                                = "Location";
        private static final String             NOTE                                    = "Body";
        //private static final String             REVISION                                = "Revision";
        private static final String             SENSITIVITY                             = "Sensitivity";                    // bugfix
        private static final String             START                                   = "Start";
        private static final String             SUMMARY                                 = "Subject";
        private static final String             ATTENDEES                               = "Attendees";
        private static final String             ATTENDEE                                = "Attendee";
        private static final String             ATTENDEE_NAME                           = "AttendeeName";
        private static final String             ATTENDEE_EMAIL                          = "AttendeeEmail";
        private static final String             ATTENDEE_STATUS                         = "AttendeeStatus";
        private static final String             ATTENDEE_TYPE                           = "AttendeeType";

        public static final String              APPOINTMENT_TYPE_KEY                    = "appointment";
/*
<Attendees>
        <Attendee>
             <AttendeeName>OneOne</AttendeeName>
             <AttendeeEmail>name@xxxxxxxxx</AttendeeEmail>
             <AttendeeStatus>1</AttendeeStatus>
             <AttendeeType>1</AttendeeType>
        </Attendee>
        <Attendee>
             <AttendeeName>TwoTwo</AttendeeName>
             <AttendeeEmail>name@xxxxxxxxx</AttendeeEmail>
             <AttendeeStatus>2</AttendeeStatus>
             <AttendeeType>2</AttendeeType>
        </Attendee>
</Attendees>
*/
    private static final int[] SIF_SUPPORTED_FIELDS =
    {
        BlackBerryEvent.START
        ,BlackBerryEvent.END
        ,BlackBerryEvent.ALLDAY
        ,BlackBerryEvent.SUMMARY
        ,BlackBerryEvent.LOCATION
        ,BlackBerryEvent.NOTE
        ,BlackBerryEvent.ALARM
        ,BlackBerryEvent.CLASS
        ,BlackBerryEvent.FREE_BUSY
    };

    
    /**
     * The constructor
     */
    public SIFEventHelper(int pastlimit, int futurelimit)
    {
        super(pastlimit,futurelimit);
    }
    
    private static Event addEvent(EventList list, Event event, Hashtable eventMap) throws PIMException
    {
        String fieldname = null;
        try
        {
            long dateStart = 0;
            boolean allday = false;
            String field = null;
            fieldname = SIFEventHelper.ALLDAY;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            field = PIMItemHelper.getValue(eventMap, SIFEventHelper.ALLDAY, false);
            if ("1".equals(field))
            {
                allday = true;
            }
            
            fieldname = SIFEventHelper.START;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.START))
            {

                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.START, false);
                if (field.length() > 0)
                {
                    if(allday)
                    {
                        Calendar cal = PIMItemHelper.parseDate(field);
                        if (PIMItemHelper.getDefaultTimeZoneOffset()>=0)
                        {
                            int recurType = SIFRepeatRuleHelper.getRecurrenceType(eventMap);
                            if ( !SIFRepeatRuleHelper.isWeeklyRecurrence(recurType) 
                                && !SIFRepeatRuleHelper.isYearlyNthRecurrence(recurType)
                                && !SIFRepeatRuleHelper.isMonthlyNthRecurrence(recurType))
                            {
                                cal.setTime(new Date(cal.getTime().getTime()+PIMItemHelper.DAY_FACTOR));
                                StaticDataHelper.info("This is an Easter TimeZone and not Weekly, MonthlyNth, YearlyNth Recurrence. Adjust 1 extra day.");
                            }
                        }
                        dateStart = PIMItemHelper.getStartOfDay(cal).getTime().getTime();
                    }
                    else
                    {
                        dateStart = PIMItemHelper.parseDateTime(field).getTime().getTime();
                    }

                    if (dateStart < 0)
                    {
                        /*
                        * if the year is before 1970, the year is set to
                        * 1970 for the BlackBerry device
                        */
                        dateStart = PIMItemHelper.adjustDateToEpoch(dateStart);
                    }
                    //final SimpleDateFormat formatter = new SimpleDateFormat(DATETIME_FORMAT_UTC);
                    event.addDate(BlackBerryEvent.START, BlackBerryEvent.ATTR_NONE, dateStart);
                    StaticDataHelper.debug("The Date that was set to StartDate ["+new Date(dateStart)+"]");
                }
            }

            fieldname = SIFEventHelper.END;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.END))
            {
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.END, false);
                if (field.length() > 0)
                {
                    long dateEnd = 0;
                    if(allday)
                    {
                        Calendar cal = PIMItemHelper.parseDate(field);
                        if (PIMItemHelper.getDefaultTimeZoneOffset()>=0)
                        {
                            int recurType = SIFRepeatRuleHelper.getRecurrenceType(eventMap);
                            if ( !SIFRepeatRuleHelper.isWeeklyRecurrence(recurType) 
                                && !SIFRepeatRuleHelper.isYearlyNthRecurrence(recurType) 
                                && !SIFRepeatRuleHelper.isMonthlyNthRecurrence(recurType))
                            {
                                cal.setTime(new Date(cal.getTime().getTime()+PIMItemHelper.DAY_FACTOR));
                                StaticDataHelper.info("This is an Easter TimeZone and not Weekly, MonthlyNth, YearlyNth Recurrence. Adjust 1 extra day.");
                            }
                        }
                        dateEnd = PIMItemHelper.getEndOfDay(cal).getTime().getTime();
                    }
                    else
                    {
                        dateEnd = PIMItemHelper.parseDateTime(field).getTime().getTime();
                    }
                        
                    if (dateEnd < 0)
                    {
                        dateEnd = PIMItemHelper.adjustDateToEpoch(dateEnd);
                    }
                    //final SimpleDateFormat formatter = new SimpleDateFormat(DATETIME_FORMAT_UTC);
                    event.addDate(BlackBerryEvent.END, BlackBerryEvent.ATTR_NONE, dateEnd);
                    StaticDataHelper.info("The Date that was set to EndDate ["+new Date(dateEnd)+"]");
                }
            }

            fieldname = SIFEventHelper.IS_RECURRING;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            field = PIMItemHelper.getValue(eventMap, SIFEventHelper.IS_RECURRING, false);// e.g. 1
            if (field.length()>0)
            {
                if (!"false".equalsIgnoreCase(field) && !"0".equals(field))
                {
                    RepeatRule rrule = new RepeatRule();
                    SIFRepeatRuleHelper.setRepeatRule(eventMap,rrule,false);
                    event.setRepeat(rrule);
                }
                else
                {
                    event.setRepeat(null);
                }
            }
            else
            {
                event.setRepeat(null);
            }


            fieldname = SIFEventHelper.ALLDAY;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.ALLDAY))
            {
                event.addBoolean(BlackBerryEvent.ALLDAY, BlackBerryEvent.ATTR_NONE, allday);
            }


            fieldname = SIFEventHelper.SUMMARY;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.SUMMARY))
            {
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.SUMMARY, true);
                if (event.countValues(BlackBerryEvent.SUMMARY)==0)
                {
                    event.addString(BlackBerryEvent.SUMMARY, BlackBerryEvent.ATTR_NONE, field);
                }
                else
                {
                    event.setString(BlackBerryEvent.SUMMARY, 0, BlackBerryEvent.ATTR_NONE, field);
                }
            }

            fieldname = SIFEventHelper.LOCATION;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.LOCATION))
            {
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.LOCATION, true);
                if (event.countValues(BlackBerryEvent.LOCATION)==0)
                {
                    event.addString(BlackBerryEvent.LOCATION, BlackBerryEvent.ATTR_NONE, field);
                }
                else
                {
                    event.setString(BlackBerryEvent.LOCATION, 0, BlackBerryEvent.ATTR_NONE, field);
                }
            }

            fieldname = SIFEventHelper.NOTE;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.NOTE))
            {
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.NOTE, true);
                event.addString(BlackBerryEvent.NOTE, BlackBerryEvent.ATTR_NONE, field);
            }
                
            /*if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.REVISION))
            {
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.REVISION, false);
                final long dateRevision = PIMItemHelper.getDate(field);
                event.addDate(BlackBerryEvent.REVISION, BlackBerryEvent.ATTR_NONE,dateRevision);
            }*/

            fieldname = SIFEventHelper.ALARM;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.ALARM))
            {
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.ALARM_ENABLE, false);
                if ("1".equals(field))
                {
                    field = PIMItemHelper.getValue(eventMap, SIFEventHelper.ALARM, false);
                    if (field.length()==0)
                    {
                        field = "0";
                    }
                    int remindTime = (Integer.parseInt(field) * 60);
                    remindTime = remindTime < 0 ? -1*remindTime : remindTime;
                    event.addInt(BlackBerryEvent.ALARM, BlackBerryEvent.ATTR_NONE, remindTime);
                }
            }
            
            // bugfix 869a
            fieldname = SIFEventHelper.SENSITIVITY;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.CLASS))
            {
                int sensitivity = 0;
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.SENSITIVITY, false);
                
                if (field.length()==0)
                {
                    field = "0";
                }
                switch (Integer.parseInt(field))
                {
                    case 0:// olNormal
                        sensitivity = Event.CLASS_PUBLIC;
                        break;
                    case 2:// olPrivate
                        sensitivity = Event.CLASS_PRIVATE;
                        break;
                    case 3:// olConfidential
                        sensitivity = Event.CLASS_CONFIDENTIAL;
                        break;
                }
                event.addInt(BlackBerryEvent.CLASS, Event.ATTR_NONE, sensitivity);
            }
                
            // bugfix 869b BusyStatus
            fieldname = SIFEventHelper.BUSY_STATUS;
            //StaticDataHelper.debug("Setting event field ["+fieldname+"]");
            if (PIMItemHelper.isSupportedField(list,BlackBerryEvent.FREE_BUSY))
            {
                field = PIMItemHelper.getValue(eventMap, SIFEventHelper.BUSY_STATUS, false);
                if (field.length()==0)
                {
                    field = ""+BlackBerryEvent.FB_BUSY;
                }
                int status = 0;
                switch (Integer.parseInt(field))
                {
                    case 0:// olFree
                        status = BlackBerryEvent.FB_FREE;
                        break;
                    case 1:// olTentative
                        status = BlackBerryEvent.FB_TENTATIVE;
                        break;
                    case 2:// olBusy
                        status = BlackBerryEvent.FB_BUSY;
                        break;
                    case 3:// olOutOfOffice
                        status = BlackBerryEvent.FB_OUT_OF_OFFICE;
                        break;
                }
                event.addInt(BlackBerryEvent.FREE_BUSY, BlackBerryEvent.ATTR_NONE, status);
            }
            return event;
        }
        catch (final Throwable e)
        {
            if (fieldname.equals(SIFEventHelper.IS_RECURRING))
            {
                fieldname = SIFRepeatRuleHelper.popSettingFieldName();
            }
            StaticDataHelper.error("SIFEventHelper.addEvent() on field ["+fieldname+"]. Exception -> "+ e.toString());
            throw new PIMException("Exception while setting value to Event field ["+fieldname+"]");
        }                
    }

    
    /**
     * Based on a property file containing all known issues related to all
     * BlackBerry devices...
     */
    private static boolean isBuggy(String knownIssue)
    {
        return true;// it has to be false
    }

    
    private static void appendXMLRecurrence(PIMList list, StringBuffer eventBuffer, Event event, String patternStartDate) throws PIMException
    {
        // enhancement 868 (recurrence)
        final RepeatRule rule = event.getRepeat();
        if (rule == null)
        {
            XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.IS_RECURRING, "0",false);
        }
        else if (patternStartDate!=null)
        {
            XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.IS_RECURRING, "1",false);
            eventBuffer.append(SIFRepeatRuleHelper.toString(rule,patternStartDate,patternStartDate,false));
        }
        else
        {
            throw new PIMException("An Event must have a StartDate to have Recurrence.");
        }
    }
    

    private static void appendXMLBusyStatus(PIMList list, StringBuffer eventBuffer, BlackBerryEvent event)
    {
        if (list.isSupportedField(BlackBerryEvent.FREE_BUSY))// 20000929
        {
            int busyStatus = 2; //olBusy
            try
            {
                switch (event.getInt(BlackBerryEvent.FREE_BUSY, 0))
                {
                    case BlackBerryEvent.FB_FREE:
                        busyStatus = 0;// olFree
                        break;
                    case BlackBerryEvent.FB_TENTATIVE:
                        busyStatus = 1;// olTentative
                        break;
                    case BlackBerryEvent.FB_BUSY:
                        busyStatus = 2;// olBusy
                        break;
                    case BlackBerryEvent.FB_OUT_OF_OFFICE:
                        busyStatus = 3;// olOutOfOffice
                        break;
                }
            }
            catch (final Exception e)
            {
                StaticDataHelper.debug("Default value was used because of exception while retrieving SIF-XML field ["+SIFEventHelper.BUSY_STATUS+"]");
            }
            XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.BUSY_STATUS, ""+busyStatus,false);
        }
    }
    
    /*
     * bugfix 869a (but Event.CLASS is unfortunately unsupported: to not pollute
     * the data coming to the server from an other client, it is important that
     * this tag is not added to the SIF-E to the server!)
     */
    private static void appendXMLSensitivity(PIMList list,StringBuffer eventBuffer, BlackBerryEvent event)
    {
        /*
        * using unsupported Event.CLASS field causes an
        * UnsupportedFieldException
        */
        if (list.isSupportedField(Event.CLASS))
        {
            int c = 0;
            try
            {
                switch (event.getInt(Event.CLASS, 0))
                {
                    case Event.CLASS_PUBLIC:// 202
                        c = 0;// <Sensitivity>0</Sensitivity>
                        break;
                    case Event.CLASS_PRIVATE:// 201
                        c = 2;// <Sensitivity>2</Sensitivity>
                        break;
                    case Event.CLASS_CONFIDENTIAL:
                        c = 3;
                        break;
                }
            }
            catch (final Exception e)
            {
                StaticDataHelper.debug("Default value was used because of exception while retrieving IF-XML field ["+SIFEventHelper.SENSITIVITY+"]");
            }
            XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.SENSITIVITY, ""+c,false);
        }
    }
    
    /**
     * This method appends the event information int Type supported by
     * blackberry to a string buffer.
     * @param PIMList: An instance of PIMList
     * @param StringBuffer: StringBuffer containing event information.
     * @param Event: Blackberry event
     * @return void.
     */
    private static void appendXMLFieldAlarmMB(PIMList list, StringBuffer eventBuffer, BlackBerryEvent event)
    {
        String value = null;
        if (list.isSupportedField(BlackBerryEvent.ALARM))
        {
            int reminder = 0;
            value = "";
            try
            {
                reminder = event.getInt(BlackBerryEvent.ALARM, 0) / 60;
                value = String.valueOf(reminder);
            }
            catch (final Exception e)
            {
                StaticDataHelper.debug("Default value was used because of exception while retrieving SIF-XML field ["+SIFEventHelper.ALARM+"]");
            }
            XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.ALARM, value,false);
            if (reminder > 0)
            {
                XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.ALARM_ENABLE, "1",false);
            }
            else
            {
                XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.ALARM_ENABLE, "0",false);
            }
        }
        else
        {
            StaticDataHelper.debug("BlackBerryEvent does not support BlackBerryEvent.ALARM");
        }
    }
    
    private static void appendXMLAttendees(PIMList list, StringBuffer eventBuffer, Event event)
    {
        if (list.isSupportedField(BlackBerryEvent.ATTENDEES))
        {
            String value = event.getString(BlackBerryEvent.ATTENDEES,0);
            StaticDataHelper.debug("BlackBerry.ATTENDEES value is ["+value+"]");
        }
    }
    
    /**
     * This method appends the event information String Type supported by blackberry to a string buffer.
     * @param PIMList: An instance of PIMList
     * @param StringBuffer: StringBuffer containing event information.
     * @param Event: Blackberry event
     * @param int: type of field
     * @param String: Name of field
     */
    private static void appendXMLField(PIMList list, StringBuffer eventBuffer, Event event, int field, String tag)
    {
        String value = null;
        if (list.isSupportedField(field))
        {
            value = "";
            try
            {
                value = event.getString(field, 0);
            }
            catch (final Exception e)
            {
                StaticDataHelper.debug("Default value was used because of exception while retriving SIF-XML field ["+tag+"]");
            }
            XmlUtil.addElementNewLine(eventBuffer, tag, value,true);
        }
        value =null;
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#toString(PIMList pimlist,PIMItem item)
     */
    public String toString(PIMList pimlist,PIMItem item) throws PIMException
    {
        if (item instanceof BlackBerryEvent)
        {
            BlackBerryEvent event = (BlackBerryEvent)item;
            final StringBuffer eventBuffer = new StringBuffer();
            eventBuffer.append(XmlUtil.createOpenTagNewLine(SIFEventHelper.APPOINTMENT_TYPE_KEY));
                
            boolean allday = PIMItemHelper.appendXMLFieldBoolean(pimlist,eventBuffer, event, BlackBerryEvent.ALLDAY,SIFEventHelper.ALLDAY); // 20000928
            //String patternStartDate = SIFEventHelper.appendXMLFieldDate(pimlist,eventBuffer, event, BlackBerryEvent.START, SIFEventHelper.START, allday); //106 "Start"
            
            String patternStartDate = "";
            Calendar StartDate = PIMItemHelper.getFieldDate(pimlist, event, BlackBerryEvent.START, false, allday, false);
            if (StartDate != null) {
                if (allday) {
                    patternStartDate = PIMItemHelper.formatAllDay(StartDate);
                }
                else {
                    patternStartDate = PIMItemHelper.formatDateTimeUTC(StartDate);
                }
            }
            StaticDataHelper.info("SIFEventHelper.toString() - START - my format ["+ patternStartDate +"]");
            XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.START, patternStartDate, false);
            

            //SIFEventHelper.appendXMLFieldDate(pimlist,eventBuffer, event, BlackBerryEvent.END, SIFEventHelper.END, allday);// 102 "End"
            
            String patternEndDate = "";
            Calendar EndDate = PIMItemHelper.getFieldDate(pimlist, event, BlackBerryEvent.END, true, allday, false);
            if (EndDate != null) {
                if (allday) {
                    patternEndDate = PIMItemHelper.formatAllDay(EndDate);
                }
                else {
                    patternEndDate = PIMItemHelper.formatDateTimeUTC(EndDate);
                }
            }
            StaticDataHelper.info("SIFEventHelper.toString() - END - my format ["+ patternEndDate +"]");
            //XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.END, patternStartDate, false);
            XmlUtil.addElementNewLine(eventBuffer, SIFEventHelper.END, patternEndDate, false);


            SIFEventHelper.appendXMLField(pimlist,eventBuffer,event, BlackBerryEvent.SUMMARY,SIFEventHelper.SUMMARY); //107
            SIFEventHelper.appendXMLField(pimlist,eventBuffer,event, BlackBerryEvent.LOCATION,SIFEventHelper.LOCATION); //103
            SIFEventHelper.appendXMLField(pimlist,eventBuffer,event, BlackBerryEvent.NOTE,SIFEventHelper.NOTE);// 104 <Body>
            SIFEventHelper.appendXMLFieldAlarmMB(pimlist,eventBuffer, event); // 100
            SIFEventHelper.appendXMLSensitivity(pimlist,eventBuffer, event); // bugfix 869a
            SIFEventHelper.appendXMLBusyStatus(pimlist,eventBuffer, event); // bugfix 869b

            
            // enhancement 868 (recurrence)
            //SIFEventHelper.appendXMLRecurrence(pimlist,eventBuffer, event,patternStartDate);
            patternStartDate = StartDate != null ? PIMItemHelper.formatAllDay(StartDate) : "";
            SIFEventHelper.appendXMLRecurrence(pimlist,eventBuffer, event, patternStartDate);
            
            eventBuffer.append(XmlUtil.createCloseTagNewLine(SIFEventHelper.APPOINTMENT_TYPE_KEY));
            return eventBuffer.toString();
        }
        return null;
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#getUID(PIMItem item)
     */
    public String getUID(PIMItem item) throws PIMException
    {
        if (item instanceof BlackBerryEvent)
        {
            try
            {
                return item.getString(BlackBerryEvent.UID,0);
            }
            catch(final Exception e)
            {
                final String msg = "Exception while retrieving UID for PIMItem of SyncSource [Calendar]";
                StaticDataHelper.error(msg);
                throw new PIMException(msg);
            }
        }
        //return null;
        final String msg = "This PIMItem of SyncSource [Calendar] is not a BlackBerryEvent.";
        StaticDataHelper.error(msg);
        throw new PIMException(msg);
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#createPIMItem(PIMList pimlist)
     */
    public PIMItem createPIMItem(PIMList pimlist) throws PIMException
    {
        if (pimlist instanceof EventList)
        {
            return ((EventList)pimlist).createEvent();
        }
        return null;
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#setPIMProperties(PIMList pimlist, PIMItem item,String pimString)
     */
    public void setPIMProperties(PIMList pimlist, PIMItem item,String pimString) throws PIMException
    {
        /*
        int length = pimString.length();
        pimString = pimString.substring(0,length-14)
                + "<Attendees>"
                + "<Attendee><AttendeeName>OneOne</AttendeeName><AttendeeEmail>name@xxxxxxxxx</AttendeeEmail><AttendeeStatus>1</AttendeeStatus><AttendeeType>1</AttendeeType></Attendee>"
                + "<Attendee><AttendeeName>TwoTwo</AttendeeName><AttendeeEmail>name@xxxxxxxxx</AttendeeEmail><AttendeeStatus>2</AttendeeStatus><AttendeeType>2</AttendeeType></Attendee>"
                + "</Attendees>"
                + "</appointment>";
        */
        StaticDataHelper.debug("pimString ["+pimString+"]");

        if (pimlist instanceof EventList)
        {
            EventList list = (EventList)pimlist;
            Vector multiValueTags = new Vector(2);
            multiValueTags.addElement(SIFRepeatRuleHelper.EXCLUDE_DATE);
            multiValueTags.addElement(SIFRepeatRuleHelper.INCLUDE_DATE);

            Vector multiHashtableValueTags = new Vector(1);
            multiHashtableValueTags.addElement(SIFEventHelper.ATTENDEE);

            Hashtable eventMap = PIMItemHelper.buildMapFromXML(SIFEventHelper.APPOINTMENT_TYPE_KEY,pimString,true,multiValueTags,null,multiHashtableValueTags);
            if (eventMap != null)
            {
                SIFEventHelper.addEvent(list,(Event)item,eventMap);
            }
            else
            {
                StaticDataHelper.info("Hashtable properties is null.");
            }
            list = null;
            multiValueTags = null;
            eventMap = null;
        }
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#find_and_deletePIMItem(PIMList pimlist, String itemKey, boolean deleteFlag)
     */
    public PIMItem findPIMItem(PIMList pimlist, String itemKey, boolean deleteFlag) throws PIMException
    {
        if (pimlist instanceof EventList)
        {
            EventList _pimlist = (EventList)pimlist;
            PIMItem item = null;
            Enumeration items = _pimlist.items();
            while(items.hasMoreElements())
            {
                item = (PIMItem)items.nextElement();
                if (itemKey.equals(this.getUID(item)))
                {
                    if (deleteFlag)
                    {
                        _pimlist.removeEvent((Event)item);
                    }
                    items = null;
                    _pimlist = null;
                    return item;
                }
            }
            item = null;
            items = null;
            _pimlist = null;
        }
        return null;
    }
    
    public Hashtable findPIMItems(PIMList pimlist, Vector itemKeys, boolean deleteFlag) throws PIMException
    {
        Hashtable ret = new Hashtable();
        if (pimlist instanceof EventList)
        {
            EventList _pimlist = (EventList)pimlist;
            PIMItem item = null;
            Enumeration items = _pimlist.items();
            while(items.hasMoreElements())
            {
                item = (PIMItem)items.nextElement();
                String uid = this.getUID(item);
                if (itemKeys.contains(uid))
                {
                    if (deleteFlag)
                    {
                        _pimlist.removeEvent((Event)item);
                    }
                    ret.put(uid,item);
                    itemKeys.removeElement(uid);
                }
            }
            item = null;
            items = null;
            _pimlist = null;
        }
        return ret;
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#clearPIMItem(PIMItem _item)
     */
    public void clearPIMItem(PIMItem _item) throws PIMException
    {
        PIMItemHelper.clearPIMItem(_item,BlackBerryEvent.UID,SIFEventHelper.SIF_SUPPORTED_FIELDS);
        ((Event)_item).setRepeat(null);
    }
    
    public void deleteAllPIMItems(PIMList pimlist) throws PIMException
    {
        if (pimlist instanceof EventList)
        {
            EventList _pimlist = (EventList)pimlist;
            PIMItem item = null;
            Enumeration items = _pimlist.items();
            while(items.hasMoreElements())
            {
                item = (PIMItem)items.nextElement();
                if (this.isInSyncRange(pimlist,item))
                {
                    _pimlist.removeEvent((Event)item);
                }
                else
                {
                    //StaticDataHelper.debug("Out of range item---------\n"+this.toString(pimlist,item));
                }
            }
            //StaticDataHelper.debug("Out of the while loop");
            item = null;
            items = null;
            _pimlist = null;
        }
    }
    
    public boolean isInSyncRange(PIMList pimlist, PIMItem _item) throws PIMException
    {
        long NOW = System.currentTimeMillis();
        //StaticDataHelper.debug("Before getting START");
        Calendar start = PIMItemHelper.getFieldDate(pimlist,_item,BlackBerryEvent.START,false,false,false);
        Calendar end = null;
        //StaticDataHelper.debug("Before getting RepeatRule");
        RepeatRule rrule = ((BlackBerryEvent)_item).getRepeat();
        if (rrule!=null)
        {
            //StaticDataHelper.debug("Before getting RepeatRule.END");
            end = SIFRepeatRuleHelper.getEndDate(rrule,false);
            if (end == null)
            {
                end = Calendar.getInstance();
                end.setTime(new Date(NOW));
            }
        }
        else
        {
            //StaticDataHelper.debug("Before getting END");
            end = PIMItemHelper.getFieldDate(pimlist,_item,BlackBerryEvent.END,true,false,false);
        }
        //StaticDataHelper.debug("Before checking value");
        if (start!=null && end!=null)
        {
            if (end.getTime().getTime()< (NOW-this.limitPast*PIMItemHelper.MONTH_FACTOR)
                || start.getTime().getTime() > (NOW+this.limitFuture*PIMItemHelper.MONTH_FACTOR))
            {
                //StaticDataHelper.debug("checking return false");
                return false;
            }
        }
        //StaticDataHelper.debug("checking return true");
        return true;
    }
    
}

