/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.source;

import com.funambol.blackberry.Configuration;
import java.util.Stack;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.microedition.pim.PIMItem;
import javax.microedition.pim.PIMException;

import net.rim.blackberry.api.pdap.BlackBerryPIMList;
import net.rim.blackberry.api.pdap.PIMListListener;

import net.rim.device.api.system.PersistentObject;
import net.rim.device.api.system.PersistentStore;
import net.rim.device.api.system.ControlledAccessException;

import com.funambol.util.StaticDataHelper;

import com.funambol.blackberry.helper.PIMItemHelper;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.spds.SyncItem;
import com.funambol.syncml.spds.SyncSource;

import com.funambol.syncml.protocol.SyncFilter;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.protocol.SyncMLStatus;

import com.funambol.blackberry.SyncClient;

import com.funambol.util.SyncListener;

import java.util.Calendar;

/**
 * Class for PIM sync management in BlackBerry devices
 */
public class BBPIMSyncSource implements PIMListListener, SyncSource
{
    private static final int        TRACKING_STORAGE_STATUS_NORMAL      = 0;
    private static final int        TRACKING_STORAGE_STATUS_ADD_ERR     = 1;
    private static final int        TRACKING_STORAGE_STATUS_DEL_ERR     = 2;
    private static final int        TRACKING_STORAGE_STATUS_UPD_ERR     = 4;
    
    public static final String      ERR_SAVE_ITEM_ON_SERVER_WARNING     = "message_not-save-on-server";
    private long persistenceStorageKey;
    private int type;
    private BlackBerryPIMList pimList;
    private SyncSourceTrackingData trackingData;
    private SyncSourceTrackingData trackingDataBackup;
    private PIMItemHelper helper;
    
    /** SyncSource configuration */
    protected SourceConfig config;

    /** Synchronization filter */
    protected SyncFilter filter;

    /** SyncMode, set by beginSync */
    protected int syncMode;

    // Item lists
    private static final Integer    NEW_FLAG    =   new Integer(1);
    private static final Integer    UPD_FLAG    =   new Integer(2);
    private static final Integer    DEL_FLAG    =   new Integer(4);
    protected Hashtable sentOutIDs;    
    private Stack delItems;
    private AutoCacheItems allItems, newItems, updItems;
    
    private int totalAdded,totalUpdated,totalDeleted;
    private int totalSent,totalOutOfRangeNotSent,totalNewSent,totalUpdatedSent,totalDeletedSent;
    private int minCacheSize,maxCacheSize;
    
    private String warningCode;
    
    /** The number of errors in a sync**/
    private int errorCount;

    /** The number of items to be sent to the server in the session */
    private int clientItemsNumber;
    
    /** The number of items that the server announced to send in the session */
    private int serverItemsNumber;

    /** The number of new items to be sent to the server in the session */
    private int clientAddItemsNumber;

    /** The number of replaced items to be sent to the server in the session */
    private int clientReplaceItemsNumber;

    /** The number of deleted items to be sent to the server in the session */
    private int clientDeleteItemsNumber;

    /** Status of the sync source summarized in an integer value. See constants
     * defined in SyncSource */
    protected int globalStatus;

    /** Listener of the sync process */
    private SyncListener listener;
    
    private SyncClient app;
    
    public BBPIMSyncSource(int type, SourceConfig config, BlackBerryPIMList list, PIMItemHelper helper, long persistenceStorageKey, SyncClient app, int mincachesize, int maxcachesize) throws SyncException
    {
    	this.type = type;
        this.pimList = list;
        this.pimList.addListener(this);
        this.helper = helper;
        this.persistenceStorageKey = persistenceStorageKey;
        this.app = app;
        this.config = config;
        this.minCacheSize = mincachesize;
        this.maxCacheSize = maxcachesize;
        
        initTrackingData();
    
        filter = null;
        syncMode = 0;
    }


    private void resetResource()
    {
        trackingDataBackup.clear();
        sentOutIDs = null;
        errorCount = 0;
        allItems = null;
        newItems = null;
        updItems = null;
        delItems = null;        

        // Init number of chages counters
        clientItemsNumber = serverItemsNumber = -1;
        clientAddItemsNumber = -1;
        clientReplaceItemsNumber = -1;
        clientDeleteItemsNumber = -1;
        
        totalAdded = 0;
        totalDeleted = 0;
        totalUpdated = 0;

        totalSent = 0;
        totalOutOfRangeNotSent = 0;
        totalNewSent = 0;
        totalUpdatedSent = 0;
        totalDeletedSent = 0;
        warningCode = null;
    }


    // This function must be called before resetResource() function
    private void stopCaching()
    {
        if (allItems != null)
        {
            allItems.stopCaching();
        }
        if (newItems != null)
        {
            newItems.stopCaching();
        }
        if (updItems != null)
        {
            updItems.stopCaching();
        }
    }


    /*
     * (non-Javadoc)
     * @see net.rim.blackberry.api.pdap.PIMListListener#itemAdded(PIMItem item)
     */
    public void itemAdded(PIMItem item)
    {
        try
        {
            final String uid = helper.getUID(item);
            StaticDataHelper.debug("SyncSource ["+config.getName()+"] added UID ["+uid+"]");
            trackingData.addNewKey(uid);
        }
        catch(final PIMException e)
        {
            trackingData.setStatus(BBPIMSyncSource.TRACKING_STORAGE_STATUS_ADD_ERR | trackingData.getStatus());
        }
        persistTrackingData();
    }


    /*
     * (non-Javadoc)
     * @see net.rim.blackberry.api.pdap.PIMListListener#itemRemoved(PIMItem item)
     */
    public void itemRemoved(PIMItem item)
    {
        try
        {
            final String uid = helper.getUID(item);
            StaticDataHelper.debug("SyncSource ["+config.getName()+"] removed UID ["+uid+"]");
            trackingData.addDeletedKey(uid);
        }
        catch(final PIMException e)
        {
            trackingData.setStatus(BBPIMSyncSource.TRACKING_STORAGE_STATUS_DEL_ERR | trackingData.getStatus());
        }
        persistTrackingData();
    }


    /*
     * (non-Javadoc)
     * @see net.rim.blackberry.api.pdap.PIMListListener#itemUpdated(PIMItem oldItem,PIMItem newItem)
     */
    public void itemUpdated(PIMItem oldItem, PIMItem newItem)
    {
        try
        {
            final String uid = helper.getUID(newItem);
            StaticDataHelper.debug("SyncSource ["+config.getName()+"] updated UID ["+uid+"]");
            trackingData.addUpdatedKey(uid);
        }
        catch(final PIMException e)
        {
            trackingData.setStatus(BBPIMSyncSource.TRACKING_STORAGE_STATUS_UPD_ERR | trackingData.getStatus());
        }
        persistTrackingData();
    }


    private void checkTrackingData()
    {
        if (trackingData.getStatus() != BBPIMSyncSource.TRACKING_STORAGE_STATUS_NORMAL)
        {
            SyncClient.showMessage(StaticDataHelper.getLanguage("message_not-send-to-server"));
            StaticDataHelper.info(StaticDataHelper.getLanguage("message_not-send-to-server"));
        }
    }


    private boolean initTrackingData() throws SyncException
    {
        try
        {
            final PersistentObject pobj = PersistentStore.getPersistentObject(persistenceStorageKey);
            if (pobj.getContents()==null)
            {
                trackingData = new SyncSourceTrackingData();
            }
            else
            {
                trackingData = (SyncSourceTrackingData)pobj.getContents();
            }
            return true;
        }
        catch(ControlledAccessException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,"The application could not access to the TrackingData for SyncSource ["+getName()+"]");
        }
    }


    private boolean persistTrackingData() throws SyncException
    {
        try
        {
            final PersistentObject pobj = PersistentStore.getPersistentObject(persistenceStorageKey);
            pobj.setContents(trackingData);
            pobj.commit();
            return true;
        }
        catch(ControlledAccessException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,"The application could not access to the TrackingData for SyncSource ["+getName()+"]");
        }
    }
    
    //----------------------------------------------------------- Public Methods
    
    /**
     * Returns the config of the source. The client can use this method
     * to obtain the config object and change some parameter. A setConfig()
     * must be called to actually change the source configuration.
     * @return the config of the source
     */
    public SourceConfig getConfig() 
    {
        return config;
    }


    /**
     * Sets the config of the source. The client can use this method
     * to change the config of the source configuration.
     * This operation should not be done while the sync is in progress.
     */
    public void setConfig(SourceConfig config) 
    {
        this.config = config;
    }
    
    //------------------------------------------------ SyncSource implementation
    /**
     * Returns the name of the source
     * @return the name of the source
     */
    public String getName() 
    {
        return config.getName();
    }

 
    /**
     * Returns the source URI
     * @return the absolute URI of the source
     */
    public String getSourceUri() 
    {
        return config.getRemoteUri();
    }

 
    /**
     * Returns the type of the source.
     * The types are defined as mime-types, for instance * text/x-vcard).
     * @return the type of the source
     */
    public String getType() 
    {
        return config.getType();
    }


    /**
     * Returns the encoding of the source.
     * The encoding can be 'b64' or 'none' only. The standard defines
     * also 'des' and '3des' but they are not implemented in this version
     * of the APIs.
     * @return the encoding of the source
     */
    public String getEncoding() 
    {
        return config.getEncoding();
    }


    /**
     * Returns the preferred sync mode of the source.
     * The preferred sync mode is the one that the SyncManager sends
     * to the server in the initialization phase. The server can respond
     * with a different alert code, to force, for instance, a slow.
     * @return the preferred sync mode for this source
     */
    public int getSyncMode() 
    {
        return config.getSyncMode();
    }

    /**
     * Returns the current filter for this SyncSource.
     */
    public SyncFilter getFilter() 
    {
        return filter;
    }
     
    /**
     * Set a new filter for this SyncSource
     */
    public void setFilter(SyncFilter filter) 
    {
        this.filter = filter;
    }
    
    /** 
     * Returns the next item of the store.
     **/
    public SyncItem getNextItem() throws SyncException 
    {
        if (app.isCancelled())  cancelSync();
        if (allItems==null)
        {
            StaticDataHelper.info("Source "+getName()+": no more items to send for slow sync");
            return null;
        }                

        SyncItem item = (SyncItem)allItems.getNextItem();
        if (item!=null)
        {
            totalSent++;
            updateSend();
            StaticDataHelper.info("Source "+getName()+": sending item "+item.getKey());
            return item;
        }
        else
        {
            StaticDataHelper.info("Source "+getName()+": no more items to send for slow sync");
            // All Items sent, we can free memory
            allItems = null;
            return null;
        }
    }
     
    /**
     * Returns the next new item of the store
     * (not yet sent to the server)
     */
    public SyncItem getNextNewItem() throws SyncException 
    {
        
        if (app.isCancelled())  cancelSync();
        if (newItems == null) 
        {
            StaticDataHelper.info("Source "+getName()+": no more new items to send");
            return null;
        }

        SyncItem item = (SyncItem)newItems.getNextItem();
        if (item!=null) 
        {
            totalSent++;
            totalNewSent++;
            updateSend();
            StaticDataHelper.info("Source "+getName()+": sending item "+item.getKey());
            return item;
        }
        else 
        {
            StaticDataHelper.info("Source "+getName()+": no more new items to send");
            // All Items sent, we can free memory
            newItems = null;
            return null;
        }
    }
     
    /**
     * Returns the first/next updated item of the store
     * (changed from the last sync)
     */
    public SyncItem getNextUpdatedItem() throws SyncException 
    {
        if (app.isCancelled())  cancelSync();
        if (updItems == null) 
        {
            StaticDataHelper.info("Source "+getName()+": no more updated items to send");
            return null;
        }

        SyncItem item = (SyncItem)updItems.getNextItem();
        if (item!=null) 
        {
            totalSent++;
            totalUpdatedSent++;
            updateSend();
            StaticDataHelper.info("Source "+getName()+": sending item "+item.getKey());
            return item;
        }
        else 
        {
            StaticDataHelper.info("Source "+getName()+": no more updated items to send");
            // All Items sent, we can free memory
            updItems = null;
            return null;
        }
    }
     
    /** 
     * Returns a SyncItem containing the key of the first/next
     * deleted item of the store (locally removed after the last sync,
     * but not yet deleted on server)
     */
    public SyncItem getNextDeletedItem() throws SyncException 
    {
        if (app.isCancelled())  cancelSync();
        if (delItems == null) 
        {
            StaticDataHelper.info("Source "+getName()+": no more deleted items to send");
            return null;
        }
        if (!delItems.empty()) 
        {
            SyncItem item = (SyncItem)delItems.pop();
            totalSent++;
            totalDeletedSent++;
            updateSend();
            StaticDataHelper.info("Source "+getName()+": sending item "+item.getKey());
            // No need to get the content here
            return item;
        }
        else 
        {
            StaticDataHelper.info("Source "+getName()+": no more deletetd items to send");
            // All Items sent, we can free memory
            delItems = null;
            return null;
        }
    } 
     
    /**
     * Return the number of changes that the client will send during the
     * session. This method, after the beginSync() call, should return
     * the number of items to be sent to the server.
     * The number of changes is computed by initXXXItems() during beginSync().
     * @return number of items to sent, or -1 if unknown
     */
    public int getClientItemsNumber() 
    {
        return clientItemsNumber;
    }
     
    /**
     * Return the number of new items (add) that the client will send during the
     * session. This method, after the beginSync() call, should return
     * the number of new items to be sent to the server.
     * The number of changes is computed by initXXXItems() during beginSync().
     * @return number of items to sent, or -1 if unknown
     */
    public int getClientAddNumber() 
    {
        return clientAddItemsNumber;
    }
     
    /**
     * Return the number of replaced items that the client will send during the
     * session. This method, after the beginSync() call, should return
     * the number of replaced items to be sent to the server.
     * The number of changes is computed by initXXXItems() during beginSync().
     * @return number of items to sent, or -1 if unknown
     */
    public int getClientReplaceNumber() 
    {
        return clientReplaceItemsNumber;
    }
     
    /**
     * Return the number of deleted items that the client will send during the
     * session. This method, after the beginSync() call, should return
     * the number of delted items to be sent to the server.
     * The number of changes is computed by initXXXItems() during beginSync().
     * @return number of items to sent, or -1 if unknown
     */
    public int getClientDeleteNumber() 
    {
        return clientDeleteItemsNumber;
    }
     
    /**
     * Return the number of changes that the server will send during the
     * session. This method, after the beginSync() call, should return
     * the number of items to be sent to the server.
     * @return number of changes from the server, or -1 if not announced.
     */
    public int getServerItemsNumber() 
    {
        return serverItemsNumber;
    }
     
    /**
     * Set the number of changes that the server will send during the
     * session. This method is called by the engine to notify the Source
     * of the number of changes announced by the server. If the server
     * does not announce the number of changes, the engine will call
     * this method with parameter -1.
     * @param number of changes from the server, or -1 if not announced.
     */
    public void setServerItemsNumber(int number) 
    {
        serverItemsNumber = number;
    }
     
    /** 
     * Return the Last Anchor for this source
     */
    public long getLastAnchor() 
    {
        return config.getLastAnchor();
    }
     
    /** 
     * Set the value of the Last Anchor for this source
     */
    public void setLastAnchor(long time) 
    {
        config.setLastAnchor(time);
    }
     
    /** 
     * Return the Next Anchor for this source
     */
    public long getNextAnchor() 
    {
        return config.getNextAnchor();
    }
     
    /** 
     * Set the value of the Next Anchor for this source
     */
    public void setNextAnchor(long time) 
    {
        config.setNextAnchor(time);
        app.getAppScreen().setLastSync(time);
    }
     
    /**
     * Called after SyncManager preparation and initialization just before start
     * the synchronization of the SyncSource.
     * @param syncMode the synchronization type: one of the values in
     *         sync4j.framework.core.AlertCode
     * @throws SyncException in case of error. This will stop the sync process
     */
    public void beginSync(int syncMode) throws SyncException 
    {
        //System.out.println("Begin sync for source- "+ getName() + " : " + Calendar.getInstance().getTime().toString());
        StaticDataHelper.info("Begin sync for source '" + getName() + "' with mode " + syncMode);
        PIMItemHelper.convertLocalToGMT(System.currentTimeMillis());
        warningCode = null;
        
        errorCount = 0;
        totalAdded = 0;
        totalDeleted = 0;
        totalUpdated = 0;
        sentOutIDs = new Hashtable();

        checkTrackingData();
        trackingDataBackup = (SyncSourceTrackingData)trackingData.clone();
        // Init lists
        switch(syncMode)
        {
            case SyncML.ALERT_CODE_SLOW:
                StaticDataHelper.info("SyncML.ALERT_CODE_SLOW");
            case SyncML.ALERT_CODE_REFRESH_FROM_CLIENT:
                StaticDataHelper.info("SyncML.ALERT_CODE_REFRESH_FROM_CLIENT");
                // A refresh from client is like a slow here
                initAllItems();
                // Init number of changes counter
                clientItemsNumber = (allItems != null) ? allItems.size() : 0 ;
                clientAddItemsNumber = (newItems != null) ? newItems.size() : 0 ;
                clientReplaceItemsNumber = (updItems != null) ? updItems.size() : 0 ;
                clientDeleteItemsNumber = (delItems != null) ? delItems.size() : 0 ;
                break;
            case SyncML.ALERT_CODE_FAST:
                StaticDataHelper.info("SyncML.ALERT_CODE_FAST");
            case SyncML.ALERT_CODE_ONE_WAY_FROM_CLIENT:
                StaticDataHelper.info("SyncML.ALERT_CODE_ONE_WAY_FROM_CLIENT");
                // A one way from client is like a fast here
                initNewItems();
                initUpdItems();
                initDelItems();
                // Init number of changes counter
                clientAddItemsNumber = (newItems != null) ? newItems.size() : 0 ;
                clientReplaceItemsNumber = (updItems != null) ? updItems.size() : 0 ;
                clientDeleteItemsNumber = (delItems != null) ? delItems.size() : 0 ;
                clientItemsNumber = clientAddItemsNumber + clientReplaceItemsNumber + clientDeleteItemsNumber;
                break;
            case SyncML.ALERT_CODE_ONE_WAY_FROM_SERVER:
                StaticDataHelper.info("SyncML.ALERT_CODE_ONE_WAY_FROM_SERVER");
                // No modifications to send (it's not
                // strictly necessary to reset the lists,
                // because the engine will not ask items to
                // the SyncSource, but it's good to do it)
                newItems = null;
                updItems = null;
                delItems = null;
                // Init number of changes counter
                clientItemsNumber = 0;
                clientAddItemsNumber = 0;
                clientReplaceItemsNumber = 0;
                clientDeleteItemsNumber = 0;
                break;
            case SyncML.ALERT_CODE_REFRESH_FROM_SERVER:
                StaticDataHelper.info("SyncML.ALERT_CODE_REFRESH_FROM_SERVER");
                // In this case, the SyncSource should
                // delete all the items in the database
                // (possibly asking the user before that)
                // No modifications to send.
                
                newItems = null;
                updItems = null;
                delItems = null;
                //newItemsCache = null;
                //updItemsCache = null;
                // Init number of changes counter
                clientItemsNumber = 0;
                clientAddItemsNumber = 0;
                clientReplaceItemsNumber = 0;
                clientDeleteItemsNumber = 0;
                try
                {
                    helper.deleteAllPIMItems(pimList);
                }
                catch(final Exception e)
                {
                    throw new SyncException(SyncException.CLIENT_ERROR,"SyncSource "+getName()+" could not delete all PIMItems. Please manually delete them before re-sync");
                }
                // The following code lines are important to ensure the client throws away 
                // all tracking data before recover from server
                trackingData.clear();
                trackingDataBackup.clear();
                persistTrackingData();
                break;
            default:
                throw new SyncException(SyncException.SERVER_ERROR,"SyncSource "+getName()+": invalid sync mode "+getSyncMode());
        }

        totalSent = 0;
        StaticDataHelper.debug("For source ["+getName()+"] There are ["+clientAddItemsNumber+"] new items should be sent");
        StaticDataHelper.debug("For source ["+getName()+"] There are ["+clientDeleteItemsNumber+"] deleted items should be sent");
        StaticDataHelper.debug("For source ["+getName()+"] There are ["+clientReplaceItemsNumber+"] updated items should be sent");
        StaticDataHelper.debug("For source ["+getName()+"] There are ["+clientItemsNumber+"] ALL items should be sent");
        
        if (app.isCancelled())  cancelSync();
        this.syncMode = syncMode;
    }
     
    /**
     * Called just before committing the synchronization process by the
     * SyncManager. The SyncSource can stop the commit phase raising an
     * exception here.
     * @throws SyncException in case of error, to stop the commit.
     */
    public void endSync() throws SyncException
    {
        //System.out.println("End sync for source- "+ getName() + " : " + Calendar.getInstance().getTime().toString());
        StaticDataHelper.info("End sync for source " + getName());
        StaticDataHelper.info(
                "For source ["+getName()+"]: \n"+
                "there were ["+totalSent+"] total items sent \n" +
                "there were ["+totalOutOfRangeNotSent+"] total OutOfRange items were not sent \n" +
                
                "there were ["+totalNewSent+"] NEW items sent \n" +
                "there were ["+totalDeletedSent+"] DELETED items sent \n" +
                "there were ["+totalUpdatedSent+"] UPDATED items sent \n" +
                
                "there were ["+totalAdded+"] total items tried to add \n" +
                "there were ["+totalDeleted+"] total items tried to delete \n" +
                "there were ["+totalUpdated+"] total items tried to update \n"
        );

        // Release resources
        resetResource();
        //trackingData.clear();
        trackingDataBackup.clear();
        persistTrackingData();
    }
     
    /**
     * Set a sync listener.
     * @param listener the listener or null to remove it
     */
    public void setListener(SyncListener listener) 
    {
        this.listener = listener;
    }
     
    /**
     * Returns the current listener (or null if not set)
     */
    public SyncListener getListener() 
    {
        return listener;
    }
     
    /**
     * Returns the status of the sync source. The status is encoded as a bit
     * mask of the STATUS_* values
     */
    public int getStatus() 
    {
        return globalStatus;
    }
     
    public void dataReceived(String date, int size)
    {
        //Do nothing here
    }

    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#addItem(com.funambol.syncml.spds.SyncItem)
     */
    public int addItem(SyncItem item) throws SyncException
    {
        //StaticDataHelper.info("Total memory free is ["+Runtime.getRuntime().freeMemory()+"] bytes");
        //StaticDataHelper.info("Total memory is ["+Runtime.getRuntime().totalMemory()+"] bytes");
        totalAdded++;
        if (app.isCancelled())  cancelSync();
        try
        {
            StaticDataHelper.trace("Before creating a blank PIMItem.");
            PIMItem pitem = helper.createPIMItem(pimList);
            if (pitem!=null)
            {
                StaticDataHelper.trace("Before calling setPIMProperties function.");
                helper.setPIMProperties(pimList,pitem,new String(item.getContent()));
                //System.out.println("Before commiting PIMItem." + Calendar.getInstance().getTime().toString());
                try {
                	pitem.commit();
                } catch (PIMException e) {
                	String error = PIMItemHelper.handlePIMException(e);
                	logError("Unable to save new item in source [" + getName() + "]: " + error);
            		return SyncMLStatus.GENERIC_ERROR;
                }
                //System.out.println("Before trying to get UID of PIMItem." + Calendar.getInstance().getTime().toString());
                String key = helper.getUID(pitem); 
                item.setKey(key);
                
                StaticDataHelper.trace("Purposfully deleting new item key: " + key);
                trackingData.deleteNewKey(key,true);
                trackingDataBackup.deleteNewKey(key);

                key = null;
                updateReceive();
                return SyncMLStatus.SUCCESS;
            }
            else
            {
                StaticDataHelper.trace("Can not create a blank PIMItem.");
            }
        }
        catch(final Exception e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,"Exception while adding new item to SyncSource ["+getName()+"] - " + e.getMessage());
        }
        logError("Unable to save new item in source [" + getName() + "]: no error");
        return SyncMLStatus.GENERIC_ERROR;
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#updateItem(com.funambol.syncml.spds.SyncItem)
     */
    public int updateItem(SyncItem item) throws SyncException
    {
        //StaticDataHelper.info("Total memory free is ["+Runtime.getRuntime().freeMemory()+"] bytes");
        //StaticDataHelper.info("Total memory is ["+Runtime.getRuntime().totalMemory()+"] bytes");
        totalUpdated++;
        if (app.isCancelled())  cancelSync();
        try
        {
            StaticDataHelper.trace("Before searching for existed PIMItem.");
            PIMItem pitem = helper.findPIMItem(pimList,item.getKey());
            if (pitem!=null)
            {
                StaticDataHelper.trace("Before clearing properties of found PIMItem.");
                helper.clearPIMItem(pitem);
                StaticDataHelper.trace("Before calling setPIMProperties functions.");
                helper.setPIMProperties(pimList,pitem,new String(item.getContent()));
                StaticDataHelper.trace("Before commiting PIMItem.");
                try {
                	pitem.commit();
                } catch (PIMException e) {
                	String error = PIMItemHelper.handlePIMException(e);
                	logError("Unable to save updated item in source [" + getName() + "]: " + error);
            		return SyncMLStatus.GENERIC_ERROR;
                }
                StaticDataHelper.trace("Before trying to get UID of PIMItem.");
                String key = helper.getUID(pitem); 
                item.setKey(key);

                StaticDataHelper.trace("Purposfully deleting updated item key: " + key);
                trackingData.deleteUpdatedKey(key,true);
                trackingDataBackup.deleteUpdatedKey(key);
                
                key = null;
                updateReceive();
                return SyncMLStatus.SUCCESS;
            }
            logError("Unable to save updated item in source [" + getName() + "]: item not found");
            return SyncMLStatus.NOT_FOUND;
        }
        catch(final PIMException e)
        {
        	String error = PIMItemHelper.handlePIMException(e);
        	logError("Exception while updating item to SyncSource [" + getName() + "]: " + error);
        	return SyncMLStatus.GENERIC_ERROR;
        }
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#deleteItem(java.lang.String)
     */
    public int deleteItem(java.lang.String item) throws SyncException
    {
        //StaticDataHelper.info("Total memory free is ["+Runtime.getRuntime().freeMemory()+"] bytes");
        //StaticDataHelper.info("Total memory is ["+Runtime.getRuntime().totalMemory()+"] bytes");
        totalDeleted++;
        if (app.isCancelled())  cancelSync();
        try
        {
            StaticDataHelper.trace("Before trying to find and delete existed PIMItem.");
            if (helper.removePIMItem(pimList,item)!=null)
            {
                StaticDataHelper.trace("Purposfully deleting deleted item key: " + item);
                trackingData.deleteDeletedKey(item,true);
                trackingDataBackup.deleteDeletedKey(item);
                
                updateReceive();
                return SyncMLStatus.SUCCESS;
            }
            logError("Unable to delete item in source [" + getName() + "]: no error");
            return SyncMLStatus.NOT_FOUND;
        }
        catch(final PIMException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,"Exception while deleting item from SyncSource ["+getName()+"]");
        }
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#getItemContent(com.funambol.syncml.spds.SyncItem)
     */
    protected SyncItem getItemContent(SyncItem item) throws SyncException
    {
        if (app.isCancelled())  cancelSync();
        try
        {
            PIMItem pitem = helper.findPIMItem(pimList,item.getKey());
            if (pitem!=null)
            {
                item.setContent(helper.toString(pimList,pitem).getBytes());
                return item;
            }
        }
        catch(final PIMException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,"Exception while retrieving content of item from SyncSource ["+getName()+"]");
        }
        return null;
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#initAllItems()
     */
    protected void initAllItems() throws SyncException
    {
        try
        {
            //allItems = new Stack();
            //allItemsCache = new Stack();
            String key = null;
            Stack temp = new Stack();
            Enumeration items = pimList.items();
            while(items.hasMoreElements())
            {
                key = helper.getUID((PIMItem)items.nextElement());
                sentOutIDs.put(key,BBPIMSyncSource.NEW_FLAG);
                //allItems.push(new SyncItem(key));
                temp.push(new SyncItem(key));
            }
            items = null;
            key = null;
            allItems = new AutoCacheItems(helper,pimList,temp,minCacheSize,maxCacheSize,getName());
        }
        catch(final PIMException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,"Exception while initializing all items of SyncSource ["+getName()+"]");
        }
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#initDelItems()
     */
    protected void initDelItems()
    {
        StaticDataHelper.debug("Start building DelItems list");
        delItems = new Stack();
        String key = null;
        while((key=trackingData.popDeletedKey())!=null)
        {
            sentOutIDs.put(key,BBPIMSyncSource.DEL_FLAG);
            delItems.push(new SyncItem(key));
            StaticDataHelper.debug("Pop key from tracking database ["+key+"]");
        }
        key = null;
        StaticDataHelper.debug("End building DelItems list");
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#initNewItems()
     */
    protected void initNewItems()
    {
        StaticDataHelper.debug("Start building NewItems list");
        //newItems = new Stack();
        //newItemsCache = new Stack();
        String key = null;
        Stack temp = new Stack();
        while((key=trackingData.popNewKey())!=null)
        {
            sentOutIDs.put(key,BBPIMSyncSource.NEW_FLAG);
            //newItems.push(new SyncItem(key));
            temp.push(new SyncItem(key));
            StaticDataHelper.debug("Pop key from tracking database ["+key+"]");
        }
        key = null;
        StaticDataHelper.debug("End building NewItems list");
        newItems = new AutoCacheItems(helper,pimList,temp,minCacheSize,maxCacheSize,getName());
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#initUpdItems()
     */
    protected void initUpdItems()
    {
        //updItems = new Stack();
        //updItemsCache = new Stack();
        StaticDataHelper.debug("Start building UpdItems list");
        String key = null;
        Stack temp = new Stack();
        while((key=trackingData.popUpdatedKey())!=null)
        {
            sentOutIDs.put(key,BBPIMSyncSource.UPD_FLAG);
            //updItems.push(new SyncItem(key));
            temp.push(new SyncItem(key));
            StaticDataHelper.debug("Pop key from tracking database ["+key+"]");
        }
        key = null;
        StaticDataHelper.debug("End building UpdItems list");
        updItems = new AutoCacheItems(helper,pimList,temp,minCacheSize,maxCacheSize,getName());
    }

    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#setItemStatus(java.lang.String key, int status)
     */
    private static final int ERR_STATUS_CLASS   =   500;
    public void setItemStatus(String key, int status)
    {
        if (status >= ERR_STATUS_CLASS)
        {
            Integer FLAG = (Integer)sentOutIDs.get(key);
            if (FLAG == BBPIMSyncSource.NEW_FLAG)
            {
                trackingData.addNewKey(key);
            }
            else if (FLAG == BBPIMSyncSource.UPD_FLAG)
            {
                trackingData.addUpdatedKey(key);
            }
            else if (FLAG == BBPIMSyncSource.DEL_FLAG)
            {
                trackingData.addDeletedKey(key);
            }
            FLAG = null;
            warningCode = BBPIMSyncSource.ERR_SAVE_ITEM_ON_SERVER_WARNING;
            StaticDataHelper.error("Could not save to server item with UID ["+key+"] for SyncSource ["+getName()+"] returned status ["+status+"]");
        }
        sentOutIDs.remove(key);
    }


    public String getWarningCode()
    {
        return warningCode;
    }


    public void restoreTrackingData()
    {
        if (trackingDataBackup != null)
        {
            trackingData.append(trackingDataBackup);
        }
    }


    private void cancelSync() throws SyncException
    {
        StaticDataHelper.info("Cancelling sync for source ["+getName()+"]");
        restoreTrackingData();
        stopCaching();
        resetResource();
        throw new SyncException(SyncException.CLIENT_ERROR, StaticDataHelper.getLanguage("message_cancelled"));
    }


    private void updateReceive()
    {
    	if (serverItemsNumber >= 0)
    	{
	        app.getAppScreen().getSyncScreen().setGlobalStatus(
	            StaticDataHelper.getLanguage("sync_receiving") + " " + (totalAdded + totalUpdated + totalDeleted) + "/" + serverItemsNumber
	        );
    	}
    	else
    	{
	        app.getAppScreen().getSyncScreen().setGlobalStatus(
	            StaticDataHelper.getLanguage("sync_receiving") + " " + (totalAdded + totalUpdated + totalDeleted)
	        );
    	}
    }


    private void updateSend()
    {
        app.getAppScreen().getSyncScreen().setGlobalStatus(
            StaticDataHelper.getLanguage("sync_sending") + " " + totalSent + "/" + clientItemsNumber
        );
    }

    
    public boolean shouldWarn()
    {
        int deleted = trackingData.countDeletedKey();
        if (type == SyncClient.SYNC_CALENDAR)
        {
        	try{
                int remainders = 0;
          		Enumeration items = pimList.items();
           		while(items.hasMoreElements())
              	{
               		remainders++;
                    items.nextElement();
            	}
                
                int total = deleted + remainders;
                StaticDataHelper.info("Considering warning user about " + deleted + " deletes when " + remainders + " items are in list");
               
                
            	//if (deleted > (.5 * remainders))
                if (Configuration.syncCalendar && (deleted > ((2*total)/3)))
             	{ // 2/3 deleted or mail account deleted bug
                	return true;
            	}
       		}
       		catch (Exception e)
       		{
       			e.printStackTrace();
     		}			
        }
        return false;
    }


    class AutoCacheItems extends Thread
    {
        private PIMItemHelper _helper;
        private BlackBerryPIMList _pimlist;
        private Stack inputSyncItemList;
        private Stack bufferedSyncItemList;
        private int _size;
        private int _maxcachesize;
        private int _mincachesize;
        private String _name;
        private int totalOutOfRangeNotSent;

        private int _cacheSize;
        private boolean stopFlag;
        private boolean _firstRun;
        private Integer totalItemsLeft;


        public AutoCacheItems(PIMItemHelper helper, BlackBerryPIMList pimlist, Stack input_itemlist, int mincachesize, int maxcachesize, String name)
        {
            _helper = helper;
            _pimlist = pimlist;
            inputSyncItemList = input_itemlist;
            _size = inputSyncItemList.size();
            _mincachesize = mincachesize;
            _maxcachesize = maxcachesize;
            _name = name;

            totalOutOfRangeNotSent = 0;
            bufferedSyncItemList = new Stack();
            _cacheSize = _maxcachesize < _size/10 ? _maxcachesize : _size/10;
            _cacheSize = _cacheSize < _mincachesize ? _mincachesize : _cacheSize;
            stopFlag = false;
            _firstRun = true;
            totalItemsLeft = new Integer(_size);
            
        }


        public int size()
        {
            return _size;
        }


        public synchronized boolean empty()
        {
            return (bufferedSyncItemList.empty() && inputSyncItemList.empty() && totalItemsLeft.intValue()==0);
        }


        public SyncItem getNextItem()
        {
            if (_firstRun)
            {
                cacheItemContent();
                _firstRun = false;
                start();
            }

            boolean wait = false;
            do
            {
                int available = 0;
                // get the number of items available immediately
                synchronized(bufferedSyncItemList)
                {
                    available = bufferedSyncItemList.size();
                }
                
                if (available>0)// There are some item to read immediately
                {
                    synchronized(bufferedSyncItemList)
                    {
                        StaticDataHelper.info("Return the first available SyncItem");
                        return (SyncItem)bufferedSyncItemList.pop();
                    }
                }
                else // There is no item to read immediately
                {
                    // get the number of items left
                    synchronized(totalItemsLeft)
                    {
                        wait = (totalItemsLeft.intValue()>0);
                    }
                    // There is no item left, return null;
                    if (!wait)
                    {
                        StaticDataHelper.info("There is no more SyncItem to read");
                        return null;
                    }
                    // There are some items left so we wait to read them
                    else
                    {
                        StaticDataHelper.info("Waiting to read the next available SyncItem");
                        try{ Thread.sleep(500); } catch(final InterruptedException e) {}
                    }
                }
            }
            while(wait);
            return null;
        }


        // This is a force to-stop. Clear all data in queue and in cache
        public synchronized void stopCaching()
        {
            StaticDataHelper.info("Trying to stop caching data items");
            stopFlag = true;
            inputSyncItemList.removeAllElements();
            bufferedSyncItemList.removeAllElements();
            totalItemsLeft = new Integer(0);
        }


        private void cacheItemContent()
        {
            try
            {
                int size = 0;
                int realsize = 0;
                synchronized(totalItemsLeft)
                {
                    totalItemsLeft = new Integer(inputSyncItemList.size());
                    realsize = totalItemsLeft.intValue();
                }
                synchronized(bufferedSyncItemList)
                {
                    size = _cacheSize - bufferedSyncItemList.size();
                }
                size = realsize < size ? realsize : size;
            
                //StaticDataHelper.info("_TotalItemLeft ["+_TotalItemLeft+"] items content for SyncSource ["+_name+"]");
                //StaticDataHelper.info("Realsize ["+realsize+"] items content for SyncSource ["+_name+"]");
                StaticDataHelper.info("Start trying to cache ["+size+"] items content for SyncSource ["+_name+"]");
                Stack keys = new Stack();
                for(int i=0;i<size;i++)
                {
                    keys.push(((SyncItem)inputSyncItemList.pop()).getKey());
                }
                
                StaticDataHelper.info("Before calling find_and_deletePIMItem function");
                Hashtable out = _helper.findPIMItems(_pimlist,keys,false);
                keys.removeAllElements();
                keys = null;
                Enumeration ks = out.keys();
                while(ks.hasMoreElements())
                {
                    String itemKey = (String)ks.nextElement();
                    PIMItem pimItem = (PIMItem)out.get(itemKey);
                    if (_helper.isInSyncRange(_pimlist,pimItem))
                    {
                        SyncItem item = new SyncItem(itemKey);
                        item.setContent(_helper.toString(_pimlist,pimItem).getBytes());
                        synchronized(bufferedSyncItemList)
                        {
                            bufferedSyncItemList.addElement(item);
                        }
                        StaticDataHelper.info("Caching SyncItem with UID ["+itemKey+"].");
                    }
                    else
                    {
                        totalOutOfRangeNotSent++;
                        StaticDataHelper.info("Out of Sync range item with UID ["+itemKey+"] will not be sent to server.");
                    }
                }
                out = null;
                ks = null;
                synchronized(totalItemsLeft)
                {
                    totalItemsLeft = new Integer(inputSyncItemList.size());
                }
            }
            catch(final Throwable e)
            {
                throw new SyncException(SyncException.CLIENT_ERROR,"Exception while caching item's content from SyncSource ["+_name+"] "+e.getMessage());
            }
        }


        public void run()
        {
            while(!stopFlag)
            {
                synchronized(totalItemsLeft)
                {
                    stopFlag = (totalItemsLeft.intValue()==0);
                }
                if (stopFlag)
                {
                    StaticDataHelper.info("There is no more data to cache. set stopFlag = true");
                }
                else 
                {
                    int size = 0;
                    synchronized(bufferedSyncItemList)
                    {
                        size = bufferedSyncItemList.size();
                    }
                    if (size<=_cacheSize/2)
                    {
                        cacheItemContent();
                    }
                    else
                    {
                        yield();
                    }
                }
            }
        }
    }


    private void logError(String message)
    {
    	StaticDataHelper.error(message);
    	errorCount++;
    }


    public int getErrorCount()
    {
    	return errorCount;
    }
}
