/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.ui;

import com.funambol.blackberry.SyncClient;
import com.funambol.util.StaticDataHelper;

import net.rim.device.api.system.Bitmap;
import net.rim.device.api.ui.component.BitmapField;
import net.rim.device.api.ui.Font;
import net.rim.device.api.ui.component.LabelField;
import net.rim.device.api.ui.container.MainScreen;
import net.rim.device.api.ui.Field;
import net.rim.device.api.ui.MenuItem;
import net.rim.device.api.ui.component.Menu;
import net.rim.device.api.ui.container.HorizontalFieldManager;
import net.rim.device.api.ui.Color;


public class SyncScreen extends MainScreen
{	
	public static final String IMAGE_SUCCESS = "icon_complete.png";
	public static final String IMAGE_WARNING = "icon_alert.png";
	public static final String IMAGE_CANCEL  = "icon_cancel.png";
	public static final String IMAGE_LOGO    = "funambol.png";

	private static final String[] IMAGE_SYNCING = 
	{
		"icon_syncing_1.png",
		"icon_syncing_2.png",
		"icon_syncing_3.png",
		"icon_syncing_4.png"
	};
	private int frame;
	
	private final SyncClient app;

	BitmapField image;
	LabelField status;
	ColoredLabelField status_calendar_label;
	ColoredLabelField status_contact_label;
	ColoredLabelField status_task_label;
	ColoredLabelField status_note_label;
	ColoredLabelField status_calendar;
	ColoredLabelField status_contact;
	ColoredLabelField status_task;
	ColoredLabelField status_note;
	ColoredLabelField server_note;
	ColoredLabelField server_task;
	ColoredLabelField server_calendar;
	ColoredLabelField server_contact;
	ColoredLabelField client_note;
	ColoredLabelField client_task;
	ColoredLabelField client_calendar;
	ColoredLabelField client_contact;
	
	LabelField status_global;
	
	Thread animation;

	private final MenuItem		backMenu		= new MenuItem
		(StaticDataHelper.getLanguage("menu_back"), 20000, 10)
		{ public void run() {
			app.hideScreen(SyncScreen.this);
		} };
	
	private final MenuItem		cancelMenu		= new MenuItem
		(StaticDataHelper.getLanguage("menu_cancel"), 20000, 10)
    	{ public void run() {
    		if (app.inSync())
    		{
        		stopAnimation();
    			app.cancelSync();
    		}
    	} };
	
	private void setIcon(String path)
	{
		if (animation == null || animation != Thread.currentThread())
		{
			StaticDataHelper.trace("SyncScreen.setIcon(): Setting icon to " + path);
		}
		synchronized(SyncClient.getEventLock())
		{
			image.setBitmap(Bitmap.getBitmapResource(path));
		}
	}
	
	/**
	 * Sets the text at the top of the sync screen
	 * @param message
	 */
	public void setGlobalStatus(String message)
	{
		synchronized(SyncClient.getEventLock())
		{
			status_global.setText(message);
		}
	}

	/**
	 * Sets the counter at the bottom of the sync screen
	 * 
	 * @param mask
	 * @param current
	 * @param max
	 */
	public void setClientCount(int mask, int current, int max)
	{
		StaticDataHelper.trace("SyncScreen.setClientCount(): Setting count to " + current + " out of " + max);
		synchronized(SyncClient.getEventLock())
		{
			ColoredLabelField field = null;
			switch (mask)
			{
				case SyncClient.SYNC_ALL:
					break;
				case SyncClient.SYNC_CALENDAR:
					field = client_calendar;
					break;
				case SyncClient.SYNC_CONTACTS:
					field = client_contact;
					break;
				case SyncClient.SYNC_TASKS:
					field = client_task;
					break;
				case SyncClient.SYNC_NOTES:
					field = client_note;
					break;
			}
			if (field == null)
				return;

			if (max <= 0)
			{
				field.setText("");// [" + current + "]");
			}
			else
			{
				field.setText("");//[" + current + "/" + max + "]");
			}
			
			if (current <= 0)
			{
				field.setColor(Color.LIGHTGREY);
			}
			else
			{
				if (max == current)
					field.setColor(Color.GREEN);
				else
					field.setColor(Color.BLACK);
			}
		}
	}
	
	/**
	 * Sets the counter at the bottom of the screen
	 * @param mask
	 * @param current
	 * @param max
	 */
	public void setServerCount (int mask, int current, int max)
	{
		StaticDataHelper.trace("SyncScreen.setClientCount(): Setting count to " + current + " out of " + max);
		synchronized(SyncClient.getEventLock())
		{
			ColoredLabelField field = null;
			switch (mask)
			{
				case SyncClient.SYNC_ALL:
					break;
				case SyncClient.SYNC_CALENDAR:
					field = server_calendar;
					break;
				case SyncClient.SYNC_CONTACTS:
					field = server_contact;
					break;
				case SyncClient.SYNC_TASKS:
					field = server_task;
					break;
				case SyncClient.SYNC_NOTES:
					field = server_note;
					break;
			}
			if (field == null)
				return;
			
			if (max <= 0)
			{
				field.setText("");// [" + current + "]");
			}
			else
			{
				field.setText("");// [" + current + "/" + max + "]");
			}
			
			if (current <= 0)
			{
				field.setColor(Color.LIGHTGREY);
			}
			else
			{
				if (max == current)
					field.setColor(Color.GREEN);
				else
					field.setColor(Color.BLACK);
			}
		}
	}
	
	/**
	 * Set the status text for a particular category
	 * 
	 * @param mask
	 * @param message
	 */
	private void setStatusText(int mask, String message)
	{
		StaticDataHelper.trace("SyncScreen.setStatusText(): Setting status for " + mask + " to " + message);
		synchronized(SyncClient.getEventLock())
		{
			if (mask == SyncClient.SYNC_ALL)
			{
				status.setText(message);
				return;
			}
	
			if ((mask & SyncClient.SYNC_CALENDAR) != 0)
			{
				status_calendar.setText(message);
			}
			if ((mask & SyncClient.SYNC_CONTACTS) != 0)
			{
				status_contact.setText(message);
			}
			if ((mask & SyncClient.SYNC_TASKS) != 0)
			{ 
				status_task.setText(message);
			}
			if ((mask & SyncClient.SYNC_NOTES) != 0)
			{ 
				status_note.setText(message);
			}
		}
	}

	/**
	 * Set the color for a category
	 * @param mask
	 * @param color
	 */
	private void setColor(int mask, int color)
	{
		synchronized(SyncClient.getEventLock())
		{
			if ((mask & SyncClient.SYNC_CALENDAR) != 0)
			{
				status_calendar.setColor(color);
				status_calendar_label.setColor(color);
			}
			if ((mask & SyncClient.SYNC_CONTACTS) != 0)
			{
				status_contact.setColor(color);
				status_contact_label.setColor(color);
			}
			if ((mask & SyncClient.SYNC_TASKS) != 0)
			{ 
				status_task.setColor(color);
				status_task_label.setColor(color);
			}
			if ((mask & SyncClient.SYNC_NOTES) != 0)
			{ 
				status_note.setColor(color);
				status_note_label.setColor(color);
			}
		}
	}

	/**
	 * Set a category as disabled
	 * @param mask
	 */
	private void setDisabled(int mask)
	{
		setColor(mask, Color.LIGHTGREY);
	}

	/**
	 * Set a category as error
	 * @param mask
	 */
	private void setError(int mask)
	{
		setColor(mask, Color.RED);
	}
	
	/**
	 * Set a category as warning
	 * @param mask
	 */
	private void setWarning(int mask)
	{
		setColor(mask, Color.ORANGERED);
	}

	/**
	 * Clear all special states of a category
	 * @param mask
	 */
	private void setNormal(int mask)
	{
		setColor(mask, Color.BLACK);
	}
	
	/**
	 * Display a failure string globally
	 * @param message
	 */
	public void displayFailure(String message)
	{
		stopAnimation();
		StaticDataHelper.trace("SyncScreen.displayFailure(message)");
		setIcon(IMAGE_WARNING);
		setStatusText(SyncClient.SYNC_ALL, message);
	}
	
	/**
	 * Display a failure string for a particular category
	 * @param mask
	 * @param message
	 */
	public void displayFailure(int mask, String message)
	{
		stopAnimation();
		StaticDataHelper.trace("SyncScreen.displayFailure(mask, message)");
		setIcon(IMAGE_WARNING);
		setStatusText(mask, message);
		setError(mask);
	}
	
	/**
	 * Display success globally
	 * @param message
	 */
	public void displaySuccess(String message)
	{
		stopAnimation();
		StaticDataHelper.trace("SyncScreen.displaySuccess(message)");
		setIcon(IMAGE_SUCCESS);
		setStatusText(SyncClient.SYNC_ALL, message);
	}
	
	/**
	 * Display success for a particular category
	 * @param mask
	 * @param message
	 */
	public void displaySuccess(int mask, String message)
	{
		stopAnimation();
		StaticDataHelper.trace("SyncScreen.displaySuccess(mask, message)");
		setIcon(IMAGE_SUCCESS);
		setStatusText(mask, message);
		setNormal(mask);
	}
	
	/**
	 * Display an error globally
	 * @param message
	 */
	public void displayError(String message)
	{
		StaticDataHelper.trace("SyncScreen.displayError(message)");
		//setIcon(IMAGE_WARNING);
		setStatusText(SyncClient.SYNC_ALL, message);
	}
	
	/**
	 * Display an error for a particular category
	 * @param mask
	 * @param message
	 */
	public void displayError(int mask, String message)
	{
		StaticDataHelper.trace("SyncScreen.displayError(mask, message)");
		//setIcon(IMAGE_WARNING);
		setStatusText(mask, message);
		setError(mask);
	}
	
	/**
	 * Display a warning globally
	 * @param message
	 */
	public void displayWarning(String message)
	{
		StaticDataHelper.trace("SyncScreen.displayWarning(message)");
		//setIcon(IMAGE_WARNING);
		setStatusText(SyncClient.SYNC_ALL, message);
	}
	
	/**
	 * Display a warning for a particular category
	 * @param mask
	 * @param message
	 */
	public void displayWarning(int mask, String message)
	{
		StaticDataHelper.trace("SyncScreen.displayWarning(mask, message)");
		//setIcon(IMAGE_WARNING);
		setStatusText(mask, message);
		setWarning(mask);
	}
	
	/**
	 * Display the current sync status globally 
	 * @param message
	 */
	public void displayStatus(String message)
	{
		startAnimation();
		StaticDataHelper.trace("SyncScreen.displayStatus(message)");
		//setIcon(IMAGE_SYNCING);
		setStatusText(SyncClient.SYNC_ALL, message);
	}
	
	/**
	 * Display the current sync status for a particualr category
	 * 
	 * @param mask
	 * @param message
	 */
	public void displayStatus(int mask, String message)
	{
		StaticDataHelper.trace("SyncScreen.displayStatus(mask, message)");
		//setIcon(IMAGE_SYNCING);
		setStatusText(mask, message);
		setNormal(mask);
	}

	/**
	 * Display a category as disabled
	 * 
	 * @param mask
	 * @param message
	 */
	public void displayDisabled(int mask, String message)
	{
		StaticDataHelper.trace("SyncScreen.displayDisabled(mask)");
		setStatusText(mask, message);
		setDisabled(mask);
	}
	
	/**
	 * Display the sync as cancelled
	 */
	public void cancelSync()
	{
		stopAnimation();
		setStatusText(SyncClient.SYNC_ALL, StaticDataHelper.getLanguage("sync_cancelled"));
		setIcon(IMAGE_CANCEL);
	}
	
	/**
	 * Begin the animation for a sync in progress
	 */
	private void startAnimation()
	{
		stopAnimation();
		if (app.inSync())
		{
			if (animation == null)
			{
				animation = new Thread()
				{
					public void run()
					{
						SyncScreen.this.runAnimation();
					}
				};
			}
		}
		
		animation.start();
	}
	
	/**
	 * Stop the animation for a sync in progress
	 */
	private void stopAnimation()
	{
		if (animation != null && animation.isAlive())
			animation.interrupt();
		animation = null;
	}
	
	/**
	 * Thread.run function for the animation
	 */
	private void runAnimation()
	{
		try{
			while (true)
			{
				if (app.inSync())
				{
					frame = (frame+1) % SyncScreen.IMAGE_SYNCING.length;
					setIcon(SyncScreen.IMAGE_SYNCING[frame]);
					Thread.sleep(500);
				}
			}
		}
		catch (Exception e) {
			return;
		}
	}
	
	/**
	 * Create the interface
	 * @param app
	 */
	public SyncScreen(SyncClient app)
	{
		CenteredManager cm = new CenteredManager();
	
		this.app = app;
		
        LabelField title = new LabelField(StaticDataHelper.getLanguage("sync_title"));
        title.setFont(Font.getDefault().derive(
        		StaticDataHelper.TITLE_FONT_STYLE,
        		StaticDataHelper.TITLE_FONT_SIZE)
        		);
        title.setPadding(
        	StaticDataHelper.TITLE_PADDING,
        	StaticDataHelper.TITLE_PADDING,
        	StaticDataHelper.TITLE_PADDING,
        	StaticDataHelper.TITLE_PADDING
        	);
        setTitle(title);
        
        Bitmap icon;
        
        icon = Bitmap.getBitmapResource(IMAGE_LOGO);
        image = new BitmapField(icon, Field.FIELD_HCENTER);
        image.setSpace(5, 5);
        cm.add(image);
        
        cm.add(new LabelField(""));
        
        status = new LabelField(
        		StaticDataHelper.getLanguage("sync_status"),
        		Field.FIELD_HCENTER);
        status.setFont(Font.getDefault().derive(Font.BOLD));
        cm.add(status);
        cm.add(new LabelField(""));
        
        HorizontalFieldManager hfm;

        hfm = new HorizontalFieldManager(Field.FIELD_HCENTER);
        status_contact_label = new ColoredLabelField(
        		StaticDataHelper.getLanguage("sync_contacts") + ": ",
        		Field.FIELD_HCENTER);
        status_contact = new ColoredLabelField(
        		"",//StaticDataHelper.getLanguage("sync_unsynchronized"),
        		Field.FIELD_HCENTER);
        client_contact = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        server_contact = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        hfm.add(status_contact_label);
        hfm.add(status_contact);
        hfm.add(client_contact);
        hfm.add(server_contact);
        cm.add(hfm);
        
        hfm = new HorizontalFieldManager(Field.FIELD_HCENTER);
        status_calendar_label = new ColoredLabelField(
        		StaticDataHelper.getLanguage("sync_calendar") + ": ",
        		Field.FIELD_HCENTER);
        status_calendar = new ColoredLabelField(
        		"",//StaticDataHelper.getLanguage("sync_unsynchronized")
        		Field.FIELD_HCENTER);
        client_calendar = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        server_calendar = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        hfm.add(status_calendar_label);
        hfm.add(status_calendar);
        hfm.add(client_calendar);
        hfm.add(server_calendar);
        cm.add(hfm);
        
        hfm = new HorizontalFieldManager(Field.FIELD_HCENTER);
        status_task_label = new ColoredLabelField(
        		StaticDataHelper.getLanguage("sync_tasks") + ": ",
        		Field.FIELD_HCENTER);
        status_task = new ColoredLabelField(
        		"",//StaticDataHelper.getLanguage("sync_unsynchronized"),
        		Field.FIELD_HCENTER);
        client_task = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        server_task = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        hfm.add(status_task_label);
        hfm.add(status_task);
        hfm.add(client_task);
        hfm.add(server_task);
        cm.add(hfm);
        
        hfm = new HorizontalFieldManager(Field.FIELD_HCENTER);
        status_note_label = new ColoredLabelField(
        		StaticDataHelper.getLanguage("sync_notes") + ": ",
        		Field.FIELD_HCENTER);
        status_note = new ColoredLabelField(
        		"",//StaticDataHelper.getLanguage("sync_unsynchronized"),
        		Field.FIELD_HCENTER);
        client_note = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        server_note = new ColoredLabelField(
        		"",// []",
        		Field.FIELD_HCENTER
        		);
        hfm.add(status_note_label);
        hfm.add(status_note);
        hfm.add(client_note);
        hfm.add(server_note);
        cm.add(hfm);
        
        cm.add(new LabelField());
        cm.add(new LabelField());
        status_global = new LabelField();
        
        cm.add(status_global);
        
        add(cm);
	}
	
	/**
	 * Create the menu
	 */
	protected void makeMenu(Menu menu, int instance)
	{
        menu.add(backMenu);
		if (app.inSync())
		{
			menu.add(cancelMenu);
		}
	}

	
}
